"""Backend-agnostisc shims."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, List, Union

import numpy as np

from bitfount.config import _PYTORCH_ENGINE, BITFOUNT_ENGINE
from bitfount.exceptions import BitfountEngineError

if TYPE_CHECKING:
    from bitfount.types import _TensorLike


class BackendTensorShim(ABC):
    """A shim for handling tensors of a particular type.

    An abstract class representing a shim/bridge for tensor handling in a particular
    backend.
    """

    @staticmethod
    @abstractmethod
    def to_numpy(t: Union[_TensorLike, List[float]]) -> np.ndarray:
        """Converts a tensor into a numpy array and returns it.

        Args:
            t: The tensor or list to convert.

        Returns:
            A numpy array.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_tensor(p: Any, **kwargs: Any) -> _TensorLike:
        """Converts the supplied argument to a tensor (if possible) and returns it.

        Args:
            p: The argument to convert to a tensor.
            **kwargs: Additional keyword arguments to pass to the tensor constructor.

        Returns:
            A tensor.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_list(p: Union[np.ndarray, _TensorLike]) -> List[float]:
        """Converts the supplied tensor or numpy array to a list and returns it.

        Args:
            p: The tensor or numpy array to convert to a list.

        Returns:
            A list.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def is_tensor(p: Any) -> bool:
        """Checks if the argument is a tensor.

        Args:
            p: The argument to check.

        Returns:
            True if the supplied argument is a tensor according to this model's
                backend, False otherwise.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def clamp_params(
        p: _TensorLike, prime_q: int, precision: int, num_workers: int
    ) -> _TensorLike:
        """Clamps the parameter of a given tensor.

        Constrains the parameters for secure sharing to be within the
        required range for secure sharing. Used only when
        `steps_between_parameter_updates is 1.`

        Args:
            p: The tensor to be constrained.
            prime_q: The prime use for secret aggregation.
            precision: The precision used for secret aggregation.
            num_workers: The number of workers taking part in the secure aggregation.

        Returns:
            The clamped parameters.
        """
        raise NotImplementedError


def _load_default_tensor_shim() -> BackendTensorShim:
    """Helper function for loading the BackendTensorShim.

    Currently only supports the PyTorchBackendTensorShim.
    """
    if BITFOUNT_ENGINE == _PYTORCH_ENGINE:
        try:
            from bitfount.backends.pytorch.federated.shim import (
                PyTorchBackendTensorShim,
            )

            return PyTorchBackendTensorShim()
        except ImportError:
            raise BitfountEngineError(
                "An error was encountered trying to load the pytorch engine; "
                "check pytorch is installed."
            )
    else:
        # Raise the same error due to only supporting shim for pytorch.
        raise BitfountEngineError(
            "An error was encountered trying to load the pytorch engine; "
            "check pytorch is installed."
        )



"""Federated early stopping."""
from dataclasses import dataclass
from typing import Dict, List

from bitfount.federated.logging import _get_federated_logger

logger = _get_federated_logger(__name__)


@dataclass
class FederatedEarlyStopping:
    """Describes a criterion for early stopping of federated model training.

    This is only applicable in the federated context where a Modeller is retrieving
    validation results from multiple workers over a training job and wants to signal
    to the workers to stop training if results are getting worse. Models already have
    their own local early stopping which is separate.

    Args:
        metric: the metric whose value is checked every iteration. Must be one
            of the metrics that is calculated by the model
        patience: number of iterations of worsening values before training is
            stopped
        delta: how much the metric needs to improve by each iteration to count
            as an improvement
    """

    metric: str
    patience: int
    delta: float

    def __post_init__(self) -> None:
        self.counter: int = 0

    def check(self, results: List[Dict[str, float]]) -> bool:
        """Checks if early stopping criteria has been met.

        Args:
            results: List of metrics

        Returns:
            True if the model training should stop training early, otherwise False.
        """
        num_results = len(results)
        if self.metric not in results[0]:
            logger.warning(
                f"Early stopping ignored. Metric {self.metric} not reported by model."
            )
        elif num_results > 1:
            one_epoch_metric_diff = results[-1][self.metric] - results[-2][self.metric]
            patience_index = -1 - self.patience if num_results > self.patience else 0
            patience_metric_diff = (
                results[-1][self.metric] - results[patience_index][self.metric]
            )
            if "loss" in self.metric:
                one_epoch_metric_diff = -one_epoch_metric_diff
                patience_metric_diff = -patience_metric_diff

            if one_epoch_metric_diff < self.delta:
                self.counter += 1
            elif patience_metric_diff >= self.delta:
                # noinspection PyAttributeOutsideInit
                #   defined in __post_init__
                self.counter = 0

        return self.counter > self.patience



"""Fixtures related to {federated,hub} helper.py functions."""
from typing import Any, Callable, Dict, Iterable
from unittest.mock import Mock, create_autospec

from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.config import BITFOUNT_STORAGE_PATH
from bitfount.hub.authentication_flow import BitfountSession


@fixture
def mock_bitfount_session(mocker: MockerFixture) -> Mock:
    """Mock out BitfountSession in the helper module."""
    # BitfountSession is used in hub creation, need to mock it out
    mock_bitfount_session: Mock = mocker.patch(
        "bitfount.hub.helper.BitfountSession",
    )
    mock_bitfount_session.return_value = create_autospec(
        BitfountSession,
        instance=True,
        # attribute needed in mock
        user_storage_path=BITFOUNT_STORAGE_PATH / "test_username",
        username="test_username",
    )
    return mock_bitfount_session


@fixture
def apply_mock_get_pod_public_keys(mocker: MockerFixture) -> Callable[[str], Mock]:
    """Allows mocking out the get_pod_public_keys function.

    Replaces it with a function that will return a dict of mock public keys.

    Returns a callable that can be used to apply the patch to the right location.
    e.g. apply_mock_get_pod_public_keys("bitfount.federated.modeller._get_pod_public_keys")  # noqa: B950
    """

    def _get_mocker_public_keys(
        pod_identifiers: Iterable[str],
        *_args: Any,
        **_kwargs: Any,
    ) -> Dict[str, Mock]:
        """Return a dict of pod identifiers to mock public keys.

        *_args and **_kwargs allows the signature to match that of the existing
        function.
        """
        return {pod_identifier: Mock() for pod_identifier in pod_identifiers}

    def apply_patch(patch_location: str) -> Mock:
        """Applies the patched method to the location specified."""
        return mocker.patch(
            patch_location,
            autospec=True,
            side_effect=_get_mocker_public_keys,
        )

    return apply_patch



"""Module containing IntermineSource class.

IntermineSource class handles loading data stored in Intermine templates.

Intermine is an open source biological data warehouse developed by the University
of Cambridge http://intermine.org/ .
The IntermineSource launches a pod that can access all templates defined under a
specified service. Please see Intermine's tutorials for a detailed overview of there
python API: https://github.com/intermine/intermine-ws-python-docs .
"""
import itertools
import logging
from typing import Any, Dict, Iterable, List, Optional, Union

from intermine.webservice import Service
import numpy as np
import pandas as pd

from bitfount.data.datasources.base_source import MultiTableSource
from bitfount.data.utils import _convert_python_dtypes_to_pandas_dtypes
from bitfount.types import _Dtypes
from bitfount.utils import delegates

logger = logging.getLogger(__name__)

INTERMINE_TYPE_MAPPING = {
    "java.lang.String": str,
    "java.lang.Double": float,
    "java.lang.Float": float,
    "java.lang.Integer": int,
    "java.lang.Boolean": bool,
    "org.intermine.objectstore.query.ClobAccess": object,
    "java.util.Date": object,
    "int": int,
}


@delegates()
class IntermineSource(MultiTableSource):
    """Data Source for loading data from Intermine templates.

    Intermine is an open source biological data warehouse developed by the University
    of Cambridge http://intermine.org/ .
    The IntermineSource launches a pod that can access all templates defined under a
    specified service. Please see Intermine's tutorials for a detailed overview of their
    python API: https://github.com/intermine/intermine-ws-python-docs.

    :::info

    You must `pip install intermine` to use this data source.

    :::
    """

    def __init__(self, service_url: str, token: Optional[str], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.service = Service(service_url, token=token)
        self.all_templates_names: Dict[
            str, List[str]
        ] = self.service.all_templates_names
        self.template_to_user_map = {
            t: user
            for user, tables in self.service.all_templates_names.items()
            for t in tables
        }
        self._check_duplicate_templates()

    def _check_duplicate_templates(self) -> None:
        """Check for duplicate template names in intermine service."""
        table_names = set()
        for t in self.table_names:
            if t in table_names:
                raise ValueError(
                    f"Duplicated template name: '{t}', found in service. "
                    "Template names must have unique names."
                )
            table_names.add(t)

    @property
    def table_names(self) -> List[str]:
        """The names of the tables accessible from this data source."""
        return list(itertools.chain(*self.all_templates_names.values()))

    def _validate_table_name(self, table_name: Optional[str] = None) -> None:
        """Validate the table name exists as a template in the Intermine service.

        Args:
            table_name: The name of the Intermine template.

        Raises:
            ValueError: If the data is multi-table but no table name provided.
            ValueError: If the table name is not found in the data.
            ValueError: If the database connection does not have any table names.
        """
        if table_name is None:
            raise ValueError("No table name provided for Intermine service.")
        elif not self.table_names:
            raise ValueError(f"Service {self.service} did not return any templates.")
        elif table_name not in self.table_names:
            raise ValueError(
                f"Template name {table_name} not found in service: {self.service}. "
                f"Available tables: {self.table_names}"
            )

    def get_values(
        self, col_names: List[str], table_name: Optional[str] = None, **kwargs: Any
    ) -> Dict[str, Iterable[Any]]:
        """Get distinct values from list of columns.

        Args:
            col_names: The list of the columns whose distinct values should be
                returned.
            table_name: The name of the table to which the column exists. Required
                for multi-table databases.

        Returns:
            The distinct values of the requested column as a mapping from col name to
            a series of distinct values.
        """
        output: Dict[str, Iterable[Any]] = {}
        if table_name:
            self._validate_table_name(table_name)

            data = self._template_to_df(table_name)
            for col_name in col_names:
                output[col_name] = data[col_name].unique()
        return output

    def get_column(
        self, col_name: str, table_name: Optional[str] = None, **kwargs: Any
    ) -> Union[np.ndarray, pd.Series]:
        """Get single column from dataset.

        Args:
            col_name: The name of the column which should be loaded.
            table_name: The name of the table to which the column exists. Required
                for multi-table databases.

        Returns:
            The column requested as a series.

        Raises:
            ValueError: If the data is multi-table but no table name provided.
            ValueError: If the table name is not found in the data.
        """
        if table_name:
            self._validate_table_name(table_name)
            data = self._template_to_df(table_name)

            return data[col_name]
        else:
            raise ValueError("Expected parameter: table_name.")

    def _template_to_df(self, template_name: str) -> pd.DataFrame:
        """Transform intermine template to pandas dataframe."""
        user = self.template_to_user_map[template_name]
        template = self.service.get_template_by_user(template_name, user)

        return pd.DataFrame(
            template.results(row="list"),
            columns=[col.replace(".", "_") for col in template.views],
        )

    def get_data(
        self, table_name: Optional[str] = None, **kwargs: Any
    ) -> Optional[pd.DataFrame]:
        """Loads and returns data from Intermine template.

        Args:
            table_name: Table name for multi table data sources. This
                comes from the DataStructure.

        Returns:
            A DataFrame-type object which contains the data.
        """
        data: Optional[pd.DataFrame] = None
        if table_name:
            self._validate_table_name(table_name)
            data = self._template_to_df(table_name)
        return data

    def get_dtypes(self, table_name: Optional[str] = None, **kwargs: Any) -> _Dtypes:
        """Loads and returns the columns and column types of the Intermine template.

        Args:
            table_name: The name of the table_name which should be loaded. Only
                required for multitable database.

        Returns:
            A mapping from column names to column types.
        """
        self._validate_table_name(table_name)
        user = self.template_to_user_map[table_name]
        template = self.service.get_template_by_user(table_name, user)
        # intermine view_types are given in their java types
        java_dtypes = dict(zip(template.views, template.view_types))
        dtypes = {
            k.replace(".", "_"): _convert_python_dtypes_to_pandas_dtypes(
                INTERMINE_TYPE_MAPPING[v], k
            )
            for k, v in java_dtypes.items()
        }
        return dtypes

    def __len__(self) -> int:
        """Intemine template length."""
        return len(self.data)

    @property
    def multi_table(self) -> bool:
        """Attribute to specify whether the datasource is multi table."""
        if len(self.table_names) > 1:
            return True
        else:
            return False



"""Fixtures related to uploading/downloading from S3."""
from datetime import datetime

from pytest import fixture

from bitfount.types import _S3PresignedPOSTFields, _S3PresignedPOSTURL, _S3PresignedURL


@fixture
def s3_netloc_subdomain() -> str:
    """A mock subdomain of the style of our message service S3 buckets."""
    time_segment = datetime.now().strftime("%Y%m%d%H%M%S%f")
    return f"-message-service-external{time_segment}.s3.eu-west-2.amazonaws.com"


@fixture
def s3_upload_url(s3_netloc_subdomain: str) -> _S3PresignedPOSTURL:
    """A faked upload URL for S3 POST."""
    return _S3PresignedPOSTURL(
        f"https://an-s3-upload-url{s3_netloc_subdomain}.s3.com/some-resource/path?hello=upload_world"  # noqa: B950  # better to keep the URL whole
    )


@fixture
def s3_upload_fields() -> _S3PresignedPOSTFields:
    """Faked upload fields for S3 POST."""
    return _S3PresignedPOSTFields({"some": "upload", "fields": "to_test_with"})


@fixture
def s3_download_url(s3_netloc_subdomain: str) -> _S3PresignedURL:
    """Fake download URL for S3 GET."""
    return _S3PresignedURL(
        f"https://an-s3-download-url{s3_netloc_subdomain}.s3.com/some-resource/path?hello=download_world"  # noqa: B950  # better to keep the URL whole
    )



"""Handling of SAML Challenges."""
import asyncio
from asyncio import Task, create_task
from typing import Awaitable, Callable, List, Mapping, MutableSequence, Optional, Tuple
import urllib.parse
import webbrowser

from aiohttp import web
from aiohttp.web import Application, AppRunner, Request, Response, TCPSite

from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.transport.identity_verification import (
    _BITFOUNT_MODELLER_PORT,
    _PORT_WAIT_TIMEOUT,
)
from bitfount.federated.transport.identity_verification.types import (
    _HasWebServer,
    _ResponseHandler,
)
from bitfount.federated.transport.message_service import _DecryptedBitfountMessage
from bitfount.federated.transport.modeller_transport import (
    _ModellerMailbox,
    _WorkerMailboxDetails,
)
from bitfount.types import _SAMLResponse

logger = _get_federated_logger(__name__)


class _SAMLWebEndpoint:
    """SAML Web Listener Endpoint.

    This is an HTTP server which listens for requests
    from the user's browser when it is redirected to
    localhost by their identity provider(s).

    As each SAML response is received they are forwarded
    on to the pod that issued the matching challenge.

    Args:
        all_saml_challenges_handled:
            An Event that will be set to complete when
            all SAML challenges have been handled.
        idp_url:
            The URL of the Identity Provider.
    """

    def __init__(
        self,
        all_saml_challenges_handled: asyncio.Event,
        idp_url: str,
    ):
        self.all_saml_challenges_handled = all_saml_challenges_handled
        self.idp_url = idp_url

        self.challenges_with_mailbox_details: MutableSequence[
            Tuple[_DecryptedBitfountMessage, _WorkerMailboxDetails]
        ] = []
        self.send: Optional[
            Callable[[_SAMLResponse, _WorkerMailboxDetails], Awaitable[None]]
        ] = None

    def set_saml_challenges(
        self,
        challenges_with_mailbox_details: MutableSequence[
            Tuple[_DecryptedBitfountMessage, _WorkerMailboxDetails]
        ],
        send: Callable[[_SAMLResponse, _WorkerMailboxDetails], Awaitable[None]],
    ) -> None:
        """Prepare SAMLWebEndpoint with challenges.

        As aiohttp can't have routes added dynamically,
        and we start up our server before we receive the SAML challenges,
        this method is needed to set the challenges and send function
        once we have received the challenges and have set up a way
        to send to the task mailboxes.

        Args:
            challenges_with_mailbox_details: SAML challenge and the
                worker mailbox details for the worker that sent it.
            send: A function for sending SAML responses to workers.
        """
        self.challenges_with_mailbox_details = challenges_with_mailbox_details
        self.send = send

    async def handle_saml_idp_response(self, request: Request) -> Response:
        """Endpoint for handling SAML responses.

        This can be seen as a recursive endpoint/function.

        Args:
            request (Request): The  aiohttp web request

        Returns:
            There are two cases:
            - Base Case (Stops the recursion):
                When? We've handled all SAML Challenges.
                Returns a Web Response showing a success message
            - Recursive Case:
                When? There's outstanding SAML Challenges
                Returns a Web Response that redirects to this
                endpoint, but with the next available SAML challenge
                Removes a SAML challenge from the list.
        """
        if self.send is None:
            logger.critical(
                "SAML Challenges & Send method were not set in SAMLWebEndpoint."
                " Unable to complete SAML authentication."
            )
            raise RuntimeError(
                "SAML Challenges & Send method were not set in SAMLWebEndpoint."
                " Unable to complete SAML authentication."
            )

        # Extract SAML response from request
        saml_response: _SAMLResponse = dict(await request.post())
        # Find where we need to send SAML response
        # And remove the challenge from the list,
        # so that next iteration we process a different SAML challenge
        _, worker_mailbox = self.challenges_with_mailbox_details.pop(0)
        logger.info(
            f"Responding to SAML challenge from: {worker_mailbox.pod_identifier}"
        )
        # Send SAML response Pod that issued challenge
        await self.send(saml_response, worker_mailbox)

        if len(self.challenges_with_mailbox_details) == 0:
            # All challenges have been processed
            # So we can allow server clean up
            # And show a message in the browser
            logger.debug(
                "All SAML Challenges from Pods handled, "
                "displaying next steps in browser."
            )
            # All challenges handled so we can set the event
            self.all_saml_challenges_handled.set()
            return Response(
                text="You've now proven your identity "
                "to all pods involved in the task. "
                "You can close this tab."
            )

        # Handle the next SAML challenge in the list by redirecting
        # The user to the IdP with the next SAML challenge
        # The IdP will redirect back to this endpoint
        next_challenge, _ = self.challenges_with_mailbox_details[0]
        saml_url = f"{self.idp_url}{urllib.parse.quote_plus(next_challenge.body)}"
        logger.info(
            f"Redirecting to the next SAML Challenge. "
            f"If something goes wrong with your authentication "
            f"then visit this URL to try again: {saml_url}"
        )
        # aiohttp uses an exception to return a redirect response
        raise web.HTTPFound(saml_url)


class _SAMLChallengeHandler(_ResponseHandler, _HasWebServer):
    """Manages SAML user authentication.

    This is used to configure and perform SAML auth
    when challenges are received from resources.

    It can start, handle & shutdown a server for
    receiving SAML authentication information.
    """

    def __init__(
        self,
        idp_url: str,
    ):
        self.idp_url = idp_url

        # This event is used to prevent us from shutting down the
        # web server until all SAML challenges have been processed
        self.all_saml_challenges_handled = asyncio.Event()

        # This task is used to ensure we don't try to
        # perform SAML authentication before the server has started
        self._server_start_task: Task

        app = Application()
        # We have to create this now so that it can be added as a route
        # As Routes cannot be added in aiohttp after the runner setup
        # has been called
        self.saml_endpoint = _SAMLWebEndpoint(
            self.all_saml_challenges_handled,
            self.idp_url,
        )
        app.add_routes(
            [web.post("/api/saml", self.saml_endpoint.handle_saml_idp_response)]
        )
        self.runner = AppRunner(app)

    def start_server(self) -> Task:
        """Sets up and starts the web server."""
        self._server_start_task = create_task(self._start())
        return self._server_start_task

    async def _start(self) -> None:
        """Sets up and starts the web server."""
        logger.debug("Starting SAML web server")
        host = "localhost"
        port = _BITFOUNT_MODELLER_PORT

        # Try to create web server for endpoint, failing out if binding to the
        # host and port isn't possible.
        try:
            await self.runner.setup()
            # This site is access by the modeller (who is running this code)
            # in their own browser.
            # Their SAML IdP provider will redirect them to:
            # `http://localhost:{BITFOUNT_MODELLER_PORT}/{PATH}`
            # So they should not have any issues accessing it
            site = TCPSite(self.runner, host, port)
            # wait_for() use ensures we don't hang indefinitely on waiting for the
            # port to be available
            await asyncio.wait_for(site.start(), _PORT_WAIT_TIMEOUT)
        except TimeoutError as e:
            logger.critical(
                f"Timeout reached whilst trying to bind SAML web endpoint to "
                f"http://{host}:{port}"
            )
            raise e
        except OSError as e:
            # Raises OSError: [Errno 48] if address already in use
            logger.critical(f"Unable to bind SAML web endpoint to http://{host}:{port}")
            raise e

        logger.debug("SAML web server started successfully")

    async def _handle_saml_challenges(
        self,
        send: Callable[[_SAMLResponse, _WorkerMailboxDetails], Awaitable[None]],
        saml_challenges: MutableSequence[_DecryptedBitfountMessage],
        worker_mailbox_details: Mapping[str, _WorkerMailboxDetails],
    ) -> None:
        """Begins the process for handling SAML Challenges.

        Updates the SAMLWebEndpoint so that it's aware of the challenges.
        Opens the first challenge in the user's browser.

        Args:
            send: Function for sending SAML response to pod.
            saml_challenges: SAML Challenges from pods.
            worker_mailbox_details: Mailbox details for each pod.
        """
        # Check that start_server() has been called
        try:
            self._server_start_task
        except AttributeError:
            raise RuntimeError(
                "SAML server has not been started; "
                "ensure start_server() has been called."
            )

        # We only open the first one, then use
        # the browser to redirect us to the other pages
        # We pair the challenges with the mailbox details
        # so that we have easy access to the keys later
        challenges_with_mailbox_details: List[
            Tuple[_DecryptedBitfountMessage, _WorkerMailboxDetails]
        ] = [
            (saml_challenge, worker_mailbox_details[saml_challenge.sender])
            for saml_challenge in saml_challenges
        ]
        self.saml_endpoint.set_saml_challenges(challenges_with_mailbox_details, send)

        if not self._server_start_task.done():
            logger.info("Waiting for SAML challenge handler to start")
            await self._server_start_task
        # If an exception was thrown in the task,
        # Task.result() will re-raise it for us.
        self._server_start_task.result()

        saml_url = f"{self.idp_url}{urllib.parse.quote_plus(saml_challenges[0].body)}"
        logger.info(
            f"Attempting to open browser. "
            f"Running a headless client? "
            f"You'll need to open this link in a browser: {saml_url}"
        )
        webbrowser.open(saml_url)

        # Wait for all SAML challenges to be handled by the listener
        await self.all_saml_challenges_handled.wait()

    async def handle(self, modeller_mailbox: _ModellerMailbox) -> None:
        try:
            # SAML is in use, so we are expecting SAML challenges from the pods
            saml_challenges = await modeller_mailbox.get_saml_challenges()

            # Process the SAML challenges & respond to them
            await self._handle_saml_challenges(
                modeller_mailbox.send_saml_responses,
                saml_challenges,
                modeller_mailbox.worker_mailboxes,
            )
        finally:
            await self.stop_server()

    async def stop_server(self) -> None:
        """Stop the web server."""
        await self.runner.cleanup()



"""General helpful utilities to be found here."""
from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from functools import reduce
import importlib
import importlib.util
import inspect
from itertools import chain
import linecache
import logging
from pathlib import Path
import random
import sys
from types import ModuleType
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Mapping,
    NoReturn,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import docstring_parser
import matplotlib.pyplot as plt
import numpy as np
from sklearn.preprocessing import OneHotEncoder

__all__: List[str] = ["one_hot_encode_list", "seed_all", "ExampleSegmentationData"]

from bitfount import config

_logger = logging.getLogger(__name__)

DEFAULT_SEED: int = 42
DOCSTRING_STYLE = docstring_parser.DocstringStyle.GOOGLE


def _is_notebook() -> bool:
    """Checks if code is being executed in a notebook or not."""
    try:
        # get_ipython() is always available in a notebook, no need to import
        shell = get_ipython().__class__.__name__  # type: ignore[name-defined] # Reason: captured by NameError # noqa: B950
        return bool(shell == "ZMQInteractiveShell")  # True if ran from Jupyter notebook
    except NameError:
        return False  # Standard Python interpreter


def _inspect_get_file(obj: Any) -> str:
    """This is a wrapper around `inspect.getfile` that works with IPython.

    This is required because `inspect.getfile` does not work with classes defined in a
    notebook due to them not having a `__file__` attribute. Functions in a notebook are
    unaffected.

    Raises:
        TypeError: If the object source code cannot be retrieved.
    """
    # Only classes need a workaround, if the object is not a class we can just fall back
    # to `inspect.getfile`.
    if not inspect.isclass(obj):
        return inspect.getfile(obj)

    # Lookup by parent module as in `inspect.getfile`. This is the flow for classes that
    # are not defined in a notebook.
    if hasattr(obj, "__module__"):
        object_ = sys.modules.get(obj.__module__)
        if object_ is not None and hasattr(object_, "__file__"):
            return str(object_.__file__)

    # If parent module is __main__, lookup by methods. This is the flow for classes that
    # are defined in a notebook.
    for _, member in inspect.getmembers(obj):
        if (
            inspect.isfunction(member)
            and obj.__qualname__ + "." + member.__name__ == member.__qualname__
        ):
            return inspect.getfile(member)

    # If all else fails, raise an error.
    raise TypeError(f"Source for {obj.__name__} not found.")


def _get_ipython_extract_symbols() -> Callable[[Any, Any], Tuple[List, List]]:
    """Returns the `extract_symbols` function from IPython.

    This should only be imported from an environment which has IPython installed.
    """
    from IPython.core.magics.code import (  # type: ignore[import] # Reason: No stubs for this # noqa: B950
        extract_symbols,
    )

    return extract_symbols  # type: ignore[no-any-return] # Reason: No stubs for this # noqa: B950


def _get_object_source_code(obj: Any) -> str:
    """Returns the source code for an object as a string.

    If the object is defined in a file, the function simply returns the result from
    `inspect.getsource`. If the object is defined in a notebook, a different approach
    must be taken to extract the source code.
    """
    if _is_notebook():
        extract_symbols = _get_ipython_extract_symbols()

        # Extract the source code from the cell the object is defined in as a string.
        cell_code = "".join(linecache.getlines(_inspect_get_file(obj)))
        # Extracts the first symbol (by name) that is defined in the cell that the
        # object is defined in. This should be the object itself.
        class_code: str = extract_symbols(cell_code, obj.__name__)[0][0]
        return class_code

    return inspect.getsource(obj)


def seed_all(seed_value: Optional[int] = None) -> None:
    """Sets random seed for `numpy`, `python.random` and any other relevant library.

    If `pytorch` is used as the backend, this is just a wrapper around the
    `seed_everything` function from `pytorch-lightning`.

    :::info

    `PYTHONHASHSEED` is not set as an environment variable because this has no effect
    once the process has been started. If you wish to set this, you should set it
    yourself in your own environment before running python.

    :::

    Args:
        seed_value: The seed value to use. If None, uses the default seed (42).
    """
    seed_value = DEFAULT_SEED if seed_value is None else seed_value

    # Try to seed other libraries that may be present
    # Torch
    try:
        import pytorch_lightning as pl

        _logger.debug(f"Setting seed of torch, random and numpy to {seed_value}")
        # This call takes care of setting the seed for all libraries including
        # random and numpy as well as pytorch. `workers=True` ensures that the workers
        # for the dataloaders are also seeded.
        pl.seed_everything(seed_value, workers=True)
    except ModuleNotFoundError:
        _logger.debug(f"Setting seed of random and numpy to {seed_value}")
        random.seed(seed_value)
        np.random.seed(seed_value)


def _add_this_to_list(
    this: Union[Any, Iterable[Any]], the_list: List[Any]
) -> List[Any]:
    """Adds this to list.

    Checks if `this` is an iterable, in which case it extends the provided list
    (unless it is a string). Otherwise it appends to the provided list if not None.

    Finally, it removes duplicates from the list and returns it.
    """
    if isinstance(this, Iterable) and not isinstance(this, str):
        the_list.extend(this)
    else:
        if this is not None:
            the_list.append(this)
    return list(set(the_list))


def _array_version(this: Union[Any, Iterable[Any]]) -> List[Any]:
    """Returns this as a list.

    Checks if 'this' is an iterable, in which case it returns it as a list (unless
    it's a string), otherwise it wraps it as a single element list.
    """
    if isinstance(this, Iterable) and not isinstance(this, str):
        return list(this)
    return [this]


def one_hot_encode_list(
    targets: Union[
        np.ndarray, Iterable[np.ndarray], Iterable[int], Iterable[Iterable[int]]
    ]
) -> np.ndarray:
    """Converts a list of targets into a list of one-hot targets."""
    arr_targets: np.ndarray = np.asarray(targets)

    # Can only encode 1D or 2D targets
    if not (1 <= arr_targets.ndim <= 2):
        raise ValueError(
            f"Incorrect number of dimensions for one-hot encoding; "
            f"expected 1 or 2, got {arr_targets.ndim}"
        )

    # OneHotEncoder needs a 2D numpy array so if this is 1D, needs reshape
    if arr_targets.ndim == 1:
        arr_targets = arr_targets.reshape(-1, 1)

    # Ensure output is ndarray of correct types
    encoder = OneHotEncoder(categories="auto", sparse=False, dtype=np.uint8)
    # sparse=False guarantees output of ndarray
    return cast(np.ndarray, encoder.fit_transform(arr_targets))


def _get_module_from_file(path_to_module: Path) -> ModuleType:
    """Imports a module from a file path and returns the module.

    Args:
        path_to_module: The path to the module file.

    Returns:
        The module.

    Raises:
        ImportError: If the module cannot be imported.
    """
    path = path_to_module.expanduser().absolute()
    module_name = path.stem

    # Load the module from the file without saving it to sys.modules
    # See: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly  # noqa: B950
    spec = importlib.util.spec_from_file_location(module_name, path)
    try:
        # `spec` has type `ModuleSpec | None`. If it is None, an error is raised which
        # caught by the `try...except` block
        module = importlib.util.module_from_spec(spec)  # type: ignore[arg-type] # Reason: See comment # noqa: B950
        spec.loader.exec_module(module)  # type: ignore[union-attr] # Reason: See comment # noqa: B950
    except Exception as ex:
        raise ImportError(f"Unable to load code from {path_to_module}") from ex

    return module


def _get_non_abstract_classes_from_module(path_to_module: Path) -> Dict[str, type]:
    """Returns non-abstract classes from a module path.

    Returns a dictionary of non-abstract class names and classes present within a
    module file which is given as a Path object.
    """
    module = _get_module_from_file(path_to_module)

    classes = {
        name: class_
        for name, class_ in vars(module).items()
        if inspect.isclass(class_) and not inspect.isabstract(class_)
    }

    return classes


# TypeVars for merge_list_of_dicts
KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


def _merge_list_of_dicts(
    lod: Iterable[Mapping[KT, Union[VT, Iterable[VT]]]]
) -> Dict[KT, List[VT]]:
    """Converts a list of dicts into a dict of lists.

    Each element in the dicts should be a single element or a list of elements.
    Any list of elements will be flattened into the final list, any single
    elements will be appended.

    Args:
        lod: A list of dicts to merge.

    Returns:
        A dictionary of lists where each value is the merged values from the
        input list of dicts.
    """
    # Merge into singular lists rather than a list of dicts
    merged: DefaultDict[KT, List[VT]] = defaultdict(list)

    for i in lod:
        for k, v in i.items():
            # We ignore the mypy errors on the `extend` and `append` calls because
            # we don't know the type of v. Instead rely on the `try...except` block
            # for type safety
            try:
                # Attempt to extend first (for outputs and targets)
                merged[k].extend(v)  # type: ignore[arg-type]  # Reason: will TypeError if not Iterable[VT]  # noqa: B950
            except TypeError:
                # Otherwise if not iterable, append
                v = cast(VT, v)
                merged[k].append(v)

    return merged


@dataclass
class _MegaBytes:
    """Thin wrapper for bytes->MB conversion."""

    whole: int
    fractional: float


def _get_mb_from_bytes(num_bytes: int) -> _MegaBytes:
    """Converts a number of bytes into the number of megabytes.

    Returns a tuple of the number of whole megabytes and the number of megabytes
    as a float.
    """
    mb: float = num_bytes / (1024 * 1024)
    return _MegaBytes(int(mb), mb)


_LOG_LEVELS_STR = Literal[
    "CRITICAL",
    "ERROR",
    "WARNING",
    "INFO",
    "DEBUG",
]
_LOG_LEVELS_NUM = Literal[
    50,  # logging.CRITICAL
    40,  # logging.ERROR
    30,  # logging.WARNING
    20,  # logging.INFO
    10,  # logging.DEBUG
]
_LOG_LEVELS = Union[_LOG_LEVELS_STR, _LOG_LEVELS_NUM]


def _find_logger_filenames(
    logger: logging.Logger, level: _LOG_LEVELS
) -> Optional[List[str]]:
    """Finds all files that are being logged to for a given logger and level."""
    # Convert to int-log-level if needed
    if isinstance(level, str):
        level = logging.getLevelName(level)
    level = cast(_LOG_LEVELS_NUM, level)

    # If logger is not enabled for target level, will never output
    if not logger.isEnabledFor(level):
        return None

    # Otherwise, iterate up through the logger hierarchy and note any FileHandler
    # files that will be output to for the target level.
    log_files: List[str] = []
    curr_logger: Optional[logging.Logger] = logger
    while curr_logger:
        handlers = curr_logger.handlers
        log_files.extend(
            h.baseFilename
            for h in handlers
            if isinstance(h, logging.FileHandler) and h.level <= level
        )
        curr_logger = curr_logger.parent  # will be None if root logger

    # Return either the list of file paths or None if no files
    # were found.
    if log_files:
        return log_files
    else:
        return None


_no_tb_limit_sentinel = object()

try:
    import gmpy2  # type: ignore[import] # Reason: see below

    # We only use 2 operations from gmpy2, and we force the output to be int
    HAVE_GMP = True
except ImportError:
    HAVE_GMP = False


def _powmod(base: int, power: int, mod: int) -> int:
    """Computes base^power modulo mod where base, power, mod are integers.

    Uses GMP if available.
    """
    if not HAVE_GMP:
        return pow(base, power, mod)
    else:
        return int(gmpy2.powmod(base, power, mod))


def _powinv(base: int, mod: int) -> int:
    """The multiplicative inverse of base in the integers modulo mod.

    Uses GMP if available.
    """
    if HAVE_GMP:
        try:
            inv = int(gmpy2.invert(base, mod))
            # According to documentation, gmpy2.invert might return 0 on
            # non-invertible element, although it seems to actually raise an
            # exception. To make sure that zero is not returned, we raise an exception
            if inv == 0:
                raise ZeroDivisionError("No inverse exists")
            return inv
        except ZeroDivisionError:
            return pow(base, -1, mod)
    else:
        return pow(base, -1, mod)


@contextmanager
def _full_traceback() -> Generator[None, None, None]:
    """Temporarily sets tracebacklimit to original limit."""
    # If sys.tracebacklimit has been set, we need to store its current value
    orig_limit: Union[int, object]
    try:
        orig_limit = sys.tracebacklimit
    except AttributeError:
        orig_limit = _no_tb_limit_sentinel

    # Reset traceback limit to default (done by deleting the attribute)
    if orig_limit is not _no_tb_limit_sentinel:
        del sys.tracebacklimit

    yield

    # Set back to original value again if needed
    if orig_limit is not _no_tb_limit_sentinel:
        sys.tracebacklimit = cast(int, orig_limit)


_fatal_logger = logging.getLogger("bitfount.fatal")


def _handle_fatal_error(
    ex: Exception,
    logger: logging.Logger = _fatal_logger,
    log_msg: Optional[str] = None,
) -> NoReturn:
    """Determines how to handle fatal errors depending on configuration.

    How the error is handled will depend on the configuration:
        - Normal, API-based use:
            - Exception is raised and system errors out.
            - Limited stack trace/error logged at CRITICAL.
            - Full stack trace logged at DEBUG.

        - Dev, API-based use:
            - Exception is raised and system errors out.
            - Full stack trace logged at CRITICAL.

        - Normal, CLI-based use:
            - Exception is captured and sys.exit() is called.
            - Limited stack trace/error logged at CRITICAL.
            - Full stack trace logged at DEBUG.

        - Dev, CLI-based use:
            - Exception is captured and sys.exit() is called.
            - Full stack trace logged at CRITICAL.

    Dev mode functionality can be configured by setting the BITFOUNT_DEV_MODE
    environment variable to True.
    CLI mode functionality can be configured by setting the BITFOUNT_CLI_MODE
    environment variable to True. This is automatically done if the Bitfount library
    is called from any of the CLI scripts provided.

    Args:
        ex: The exception to handle.
        logger: The logger to use for logging out details. If not provided a
            "bitfount.fatal" logger is used.
        log_msg: The log message to use. If not provided the str() representation
            of the exception is used.
    """
    # Create log_msg from exception if one not explicitly provided
    if not log_msg:
        log_msg = str(ex)

    # Determine how full stack trace will be logged
    if not config.BITFOUNT_DEV_MODE:
        # Will want to say "see debug log for more details"
        additional_log_msg = ""
        if debug_log_files := _find_logger_filenames(logger, "DEBUG"):
            if not log_msg.endswith("."):
                additional_log_msg += "."
            debug_log_files_str = ", ".join(debug_log_files)
            additional_log_msg += f" Full details logged to {debug_log_files_str}."
        logger.critical(log_msg + additional_log_msg)
        with _full_traceback():
            logger.debug(log_msg, exc_info=ex)
    else:
        with _full_traceback():
            logger.critical(log_msg, exc_info=ex)

    if config._BITFOUNT_CLI_MODE:
        # Exit instead of raising exception
        sys.exit(ex)
    else:
        # Exceptions get raised
        raise ex


# Example segmentation image dataset class
class ExampleSegmentationData:
    """A synthetic segmentation dataset example.

    This class is used mainly in Tutorial 11.
    """

    def masks_to_colorimg(self, masks: np.ndarray) -> np.ndarray:
        """Adds RGB coloring to masks.

        Args:
            masks: The masks to which we want to add coloring.

        Returns:
              The masks with RGB coloring applied to them.
        """
        colors = np.asarray(
            [(201, 58, 64), (242, 207, 1)]
        )  # , (0, 152, 75)])#, (101, 172, 228),(56, 34, 132), (160, 194, 56)])

        colorimg = np.ones((masks.shape[1], masks.shape[2], 3), dtype=np.float32) * 255
        channels, height, width = masks.shape
        for y in range(height):
            for x in range(width):
                selected_colors = colors[masks[:, y, x] > 0.5]

                if len(selected_colors) > 0:
                    colorimg[y, x, :] = np.mean(selected_colors, axis=0)

        return colorimg.astype(np.uint8)

    def generate_data(
        self, height: int, width: int, count: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Generate random data given height, weight and count.

        Args:
            height: The height of the images to generate.
            width: The width of the images to generate.
            count: The number of images to generate.
        """
        x, y = zip(
            *[self._generate_img_and_mask(height, width) for i in range(0, count)]
        )

        X = np.asarray(x) * 255
        X = X.repeat(3, axis=1).transpose([0, 2, 3, 1]).astype(np.uint8)
        Y = np.asarray(y)
        return X, Y

    def _generate_img_and_mask(
        self, height: int, width: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Creates image and mask given height and width."""
        shape = (height, width)
        # Assign random location to triangle and circle
        triangle_location = self._get_random_location(*shape)
        circle_location2 = self._get_random_location(*shape, zoom=1.5)

        # Create input image
        arr = np.zeros(shape, dtype=bool)
        arr = self._add_triangle(arr, *triangle_location)
        arr = self._add_circle(arr, *circle_location2, fill=True)
        arr = np.reshape(arr, (1, height, width)).astype(np.float32)

        # Create target masks
        masks = np.asarray(
            [
                self._add_circle(
                    np.zeros(shape, dtype=bool), *circle_location2, fill=True
                ),
                self._add_triangle(np.zeros(shape, dtype=bool), *triangle_location),
            ]
        ).astype(np.float32)
        return arr, masks

    def _get_random_location(
        self, height: int, width: int, zoom: float = 1.0
    ) -> Tuple[int, int, int]:
        """Gets a random location for placing shapes."""
        x = int(
            width * random.uniform(0.1, 0.9)  # nosec [B311: random] # reason: see below
        )
        y = int(
            height
            * random.uniform(0.1, 0.9)  # nosec [B311: random] # reason: see below
        )
        size = int(
            min(width, height)
            * random.uniform(0.06, 0.12)  # nosec [B311: random] # reason: see below
            * zoom
        )
        # We only use random for generating a images for a demo segmentation
        # dataset used in testing an as an example in tutorial 11
        return (x, y, size)

    def _add_triangle(self, arr: np.ndarray, x: int, y: int, size: int) -> np.ndarray:
        """Adds a triangle to the input arrays."""
        s = int(size / 2)
        triangle = np.tril(np.ones((size, size), dtype=bool))
        arr[
            x - s : x - s + triangle.shape[0], y - s : y - s + triangle.shape[1]
        ] = triangle
        return arr

    def _add_circle(
        self, arr: np.ndarray, x: int, y: int, size: int, fill: bool = False
    ) -> np.ndarray:
        """Adds a circle to the input arrays."""
        xx, yy = np.mgrid[: arr.shape[0], : arr.shape[1]]
        circle = np.sqrt((xx - x) ** 2 + (yy - y) ** 2)
        new_arr = np.logical_or(
            arr,
            np.logical_and(circle < size, circle >= size * 0.7 if not fill else True),
        )
        return new_arr

    def _plot_img_array(self, img_array: np.ndarray, ncol: int = 3) -> None:
        """Helper function for tutorial 11."""
        nrow = len(img_array) // ncol

        f, plots = plt.subplots(
            nrow, ncol, sharex="all", sharey="all", figsize=(ncol * 4, nrow * 4)  # type: ignore[arg-type] # Reason: see below # noqa: B950
        )
        # From matplotlib documentation:
        # sharex, sharey: bool or {'none', 'all', 'row', 'col'}, default: False
        for i in range(len(img_array)):
            plots[i // ncol, i % ncol]  # type: ignore[index] # reason: see below # noqa: B950
            plots[i // ncol, i % ncol].imshow(img_array[i])  # type: ignore[index] # reason: see below # noqa: B950
        # matplotlib allows indexing of `Axes` objects,
        # not sure why mypy raises this error

    def plot_side_by_side(self, img_arrays: np.ndarray) -> None:
        """Helper function for tutorial 11.

        Plots side by side images and masks.

        Args:
            img_arrays: The array of images and masks pairs to plot.
        """
        flatten_list = reduce(lambda x, y: x + y, zip(*img_arrays))

        self._plot_img_array(np.array(flatten_list), ncol=len(img_arrays))


def delegates(
    keep: bool = False, exclude_doc_meta: Optional[List[Any]] = None
) -> Callable[[Any], None]:
    """Decorator to replace `**kwargs` in signature.

    The `@delegates()` decorator should be added to all classes that inherit
    from a parent class. It will ensure that the signature of the class
    includes all the arguments from the parent class(es) as well as combine
    the docstring of all the parent classes to make sure all args, attributes
    and errors are reflected in the class it decorates.
    """

    def _f(f: Any) -> Any:
        # Mypy complains that DocstringMeta is not iterable, when it is,
        # hence the "type: ignore"s.
        # get the signature of the decorated class
        from_f = f.__init__
        sig = inspect.signature(from_f)
        sigd = dict(sig.parameters)
        # get the docstring of the decorated class
        fdoc = docstring_parser.parse(f.__doc__, style=DOCSTRING_STYLE)
        # aux variables to keep track of everything
        # "combined" keeps track of all args, attributes and errors across all classes.
        combined: Dict[Any, Any] = {}
        # "metas" keeps track of all args, attributes and errors for one class only.
        metas: Dict[Any, Any] = {}
        # "arg_names" keeps track of all the arg (also attrs & errors)
        # names to ensure there are no duplicates.
        arg_names: List[str] = []
        # get all the metadata from the decorated class' docstring
        for meta in fdoc.meta:
            meta_type = type(meta)
            if exclude_doc_meta is not None:
                if meta_type in exclude_doc_meta:
                    continue
            metas.setdefault(meta_type, []).append(meta)
        for (meta_type, meta) in metas.items():
            combined[meta_type] = meta
            arg_names += [arg.args[1] for arg in meta]  # type: ignore[attr-defined] # Reason: see top of function # noqa: B950
        if "kwargs" in sigd:
            k = sigd.pop("kwargs")
        else:
            k = None

        # now we loop through all the base classes of the
        # decorated class and extract the relevant items
        for item in f.__mro__:

            to_f = item.__init__
            # get the docstring and the metadata
            item_doc = docstring_parser.parse(item.__doc__, style=DOCSTRING_STYLE)
            # "metas" keeps track of all args, attributes and errors for one class only.
            metas = {}
            for meta in item_doc.meta:
                meta_type = type(meta)
                if exclude_doc_meta is not None:
                    if meta_type in exclude_doc_meta:
                        continue
                metas.setdefault(meta_type, []).append(meta)
            for (meta_type, meta) in metas.items():
                to_add = [arg for arg in meta if arg.args[1] not in arg_names]  # type: ignore[attr-defined] # Reason: see top of function  # noqa: B950
                arg_names += [arg.args[1] for arg in meta]  # type: ignore[attr-defined] # Reason: see top of function  # noqa: B950
                if meta_type in combined.keys():
                    combined[meta_type] += to_add
                else:
                    combined[meta_type] = to_add
            # Now get the signature.
            # extract the "self" param from the signature
            self_param = sigd.pop("self")
            # get the values without default first
            s1 = {
                k: v
                for k, v in inspect.signature(to_f).parameters.items()
                if v.default == inspect.Parameter.empty
                and k not in sigd
                and k not in ["args", "kwargs"]
            }
            # "self" needs to go first, then unparametrised args
            # and then all the other args otherwise an error is raised.
            sigd = {"self": self_param, **s1, **sigd}
            # Then add the params which have default values.
            s2 = {
                k: v
                for k, v in inspect.signature(to_f).parameters.items()
                if v.default != inspect.Parameter.empty and k not in sigd
            }

            sigd.update(s2)
        if keep and k is not None:
            sigd["kwargs"] = k
        # replace the decorated class's signature.
        from_f.__signature__ = sig.replace(parameters=sigd.values())  # type: ignore[arg-type] # Reason: see below # noqa: B950
        # Mypy complains that Argument "parameters" to "replace" of
        # "Signature" has incompatible type "dict_values[str, Parameter]";
        # expected "Union[Sequence[Parameter], Type[_void], None], even though
        # Signature.replace is untyped.

        # now sort alphabetically the args, attrs and errors in the docstring.
        for k in combined.keys():
            combined[k] = [
                item for item in sorted(combined[k], key=lambda d: d.args[1])  # type: ignore[no-any-return] # Reason: see below # noqa: B950
            ]
        # Mypy raises error for the sorted function due to the definition of sorted
        # having changed, see here: https://github.com/python/mypy/issues/9656

        # Now, replace the decorated class's docstring metadata.
        fdoc.meta = list(chain(*combined.values()))
        f.__doc__ = docstring_parser.compose(fdoc, style=DOCSTRING_STYLE)
        return f

    return _f



"""Tests PyTorch NeuralNetwork-based classes."""
import re
from typing import List, Optional, Tuple, cast

from _pytest.logging import LogCaptureFixture
import pytest
from pytest import fixture
from pytest_lazyfixture import lazy_fixture
import torch
from torch.nn import Embedding

from bitfount.backends.pytorch.models.nn import (
    _get_torchvision_classification_model,
    _PyTorchLogisticRegression,
)
from tests.utils.helper import (
    backend_test,
    get_warning_logs,
    integration_test,
    unit_test,
)


@backend_test
@unit_test
class TestPyTorchLogisticRegression:
    """Tests for _PyTorchLogisticRegression model."""

    @fixture
    def embedding_sizes(self) -> List[Tuple[int, int]]:
        """A set of embedding sizes for categorical features."""
        return [(5, 10), (10, 5), (3, 4)]

    def test_init_prioritises_input_dim(
        self, caplog: LogCaptureFixture, embedding_sizes: List[Tuple[int, int]]
    ) -> None:
        """Tests input_dim given preference over embedding args.

        When both are supplied, the embedding-related args should be ignored in
        preference of using input_dim.
        """
        log_reg = _PyTorchLogisticRegression(
            num_classes=5,
            input_dim=10,
            num_continuous=5,
            embedding_sizes=embedding_sizes,
        )

        # Check warning logs
        warning_logs = get_warning_logs(caplog)
        assert (
            "Cannot specify both input_dim and either of num_continuous or "
            "embedding_sizes. Will default to using input_dim." in warning_logs
        )

        # Check correct attributes are set
        assert log_reg.input_dim == 10
        assert log_reg.num_continuous is None
        assert log_reg.embedding_sizes is None

    def test_init_fails_if_neither_input_dim_or_embeddings_supplied(self) -> None:
        """Test init raises error if neither input_dim or embeddings supplied."""
        with pytest.raises(
            ValueError,
            match=re.escape("One of input_dim or embedding_sizes must be provided."),
        ):
            _PyTorchLogisticRegression(num_classes=5)

    @pytest.mark.parametrize(
        argnames=("embedding_sizes_arg", "num_continuous_arg"),
        argvalues=(
            pytest.param(
                lazy_fixture("embedding_sizes"), None, id="only embedding_sizes"
            ),
            pytest.param(None, 5, id="only num_continuous"),
        ),
    )
    def test_init_fails_if_mismatch_between_embedding_args(
        self,
        embedding_sizes_arg: Optional[List[Tuple[int, int]]],
        num_continuous_arg: Optional[int],
    ) -> None:
        """Test init raises error if only one embedding argument is provided.

        In particular, checks that embedding_sizes and num_continuous must both
        be provided or that neither must be.
        """
        # The error message will differ depending on which one is missing (no
        # embedding_sizes will get caught because there's also no input_dim) but
        # both cases should fail.
        with pytest.raises(ValueError):
            _PyTorchLogisticRegression(
                num_classes=5,
                embedding_sizes=embedding_sizes_arg,
                num_continuous=num_continuous_arg,
            )

    def test_init_when_using_embedding_args(
        self,
        embedding_sizes: List[Tuple[int, int]],
    ) -> None:
        """Tests init functions correctly when supplied embedding args.

        Checks that:
            - input_dim is created from supplied embedding-related args.
            - category embeddings are created
            - dropout is created
            - linear layer has correct sizes
        """
        num_classes = 10
        num_continuous = 5
        expected_input_dim = num_continuous + sum(size for _, size in embedding_sizes)

        log_reg = _PyTorchLogisticRegression(
            num_classes=num_classes,
            num_continuous=num_continuous,
            embedding_sizes=embedding_sizes,
        )

        # Check input_dim
        assert log_reg.input_dim == expected_input_dim
        # Check category embeddings
        assert log_reg.category_embeddings is not None
        for idx, embedding in enumerate(log_reg.category_embeddings):
            embedding = cast(Embedding, embedding)
            num_categories, category_embed_size = embedding_sizes[idx]
            assert embedding.num_embeddings == num_categories
            assert embedding.embedding_dim == category_embed_size
        # Check category embeddings dropout
        assert log_reg.category_embeddings_dropout is not None
        assert log_reg.category_embeddings_dropout.p == log_reg.embedding_dropout_frac
        # Check size of linear layer
        assert log_reg.linear.in_features == expected_input_dim
        assert log_reg.linear.out_features == num_classes

    def test_init_when_using_input_dim(self) -> None:
        """Tests init functions correctly when supplied input_dim.

        Checks that:
            - category embeddings are NOT created
            - dropout is NOT created
            - linear layer has correct sizes
        """
        num_classes = 10
        input_dim = 32

        log_reg = _PyTorchLogisticRegression(
            num_classes=num_classes,
            input_dim=input_dim,
        )

        # Check category embeddings
        assert log_reg.category_embeddings is None
        # Check category embeddings dropout
        assert log_reg.category_embeddings_dropout is None
        # Check size of linear layer
        assert log_reg.linear.in_features == input_dim
        assert log_reg.linear.out_features == num_classes

    @pytest.mark.parametrize(
        argnames=("embedding_sizes_arg", "num_continuous_arg"),
        argvalues=(
            pytest.param([], 10, id="continuous only"),
            pytest.param([(10, 20), (5, 15)], 0, id="categorical only"),
            pytest.param([(10, 20), (5, 15)], 10, id="categorical and continuous"),
        ),
    )
    def test_forward(
        self,
        embedding_sizes_arg: List[Tuple[int, int]],
        num_continuous_arg: int,
    ) -> None:
        """Test that forward works with various types of batch.

        Uses categorical embedding.

        Tests:
            - mixed categorical and continuous
            - categorical only
            - continuous only
        """
        # Construct appropriately shaped tensors
        batch_size = 32

        num_categorical = len(embedding_sizes_arg)
        categorical_tensor = torch.randint(
            low=0, high=5, size=(batch_size, num_categorical)
        )

        continuous_tensor = torch.rand(size=(batch_size, num_continuous_arg))

        # Categorical tensor must be transposed to allow it to pass through the
        # embedding layer; this is how it will be returned from
        # _split_dataloader_output.
        x = (categorical_tensor.t(), continuous_tensor)

        log_reg = _PyTorchLogisticRegression(
            num_classes=5,
            num_continuous=num_continuous_arg,
            embedding_sizes=embedding_sizes_arg,
        )

        # We don't assert anything on the results, we just want to check it
        # handles it without crashing.
        log_reg.forward(x)

    @pytest.mark.parametrize(
        argnames=("num_categorical_arg", "num_continuous_arg"),
        argvalues=(
            pytest.param(0, 10, id="continuous only"),
            pytest.param(5, 0, id="categorical only"),
            pytest.param(5, 10, id="categorical and continuous"),
        ),
    )
    def test_forward_without_categorical_embedding(
        self,
        num_categorical_arg: int,
        num_continuous_arg: int,
    ) -> None:
        """Test that forward works with various types of batch.

        Uses raw tensor embedding for categorical features.

        Tests:
            - mixed categorical and continuous
            - categorical only
            - continuous only
        """
        # Construct appropriately shaped tensors
        batch_size = 32

        categorical_tensor = torch.randint(
            low=0, high=5, size=(batch_size, num_categorical_arg)
        )

        continuous_tensor = torch.rand(size=(batch_size, num_continuous_arg))

        # Categorical tensor must NOT be transposed to allow it to pass through
        # the embedding layers. This is how it would be returned from
        # _split_dataloader_output.
        x = (categorical_tensor, continuous_tensor)

        log_reg = _PyTorchLogisticRegression(
            num_classes=5,
            input_dim=num_continuous_arg + num_categorical_arg,
        )

        # We don't assert anything on the results, we just want to check it
        # handles it without crashing.
        log_reg.forward(x)


@backend_test
@integration_test
class TestTorchvisionClassificationModels:
    """Tests retrieval of existing Torchvision models."""

    def test_unimplemented_model(self) -> None:
        """Tests error thrown if model exists but not supported."""
        with pytest.raises(
            ValueError,
            match="Model reshaping not implemented yet. Choose another model.",
        ):
            _get_torchvision_classification_model("googlenet", False, 2)

    def test_unrecognised_model(self) -> None:
        """Tests error thrown if model name not recognized."""
        with pytest.raises(ValueError, match="Model name not recognised"):
            _get_torchvision_classification_model("blahblahmodel", False, 2)

    def test_resnet(self) -> None:
        """Tests resnet* retrieval."""
        model = _get_torchvision_classification_model("resnet18", False, 2)
        assert model.fc.out_features == 2  # type: ignore[union-attr]  # Reason: test will fail if wrong type  # noqa: B950

    def test_alexnet(self) -> None:
        """Tests alexnet retrieval."""
        model = _get_torchvision_classification_model("alexnet", False, 2)
        assert model.classifier[6].out_features == 2  # type: ignore[index,union-attr]  # Reason: test will fail if wrong type  # noqa: B950

    def test_vgg(self) -> None:
        """Tests vgg* retrieval."""
        model = _get_torchvision_classification_model("vgg16", False, 2)
        assert model.classifier[6].out_features == 2  # type: ignore[index,union-attr]  # Reason: test will fail if wrong type  # noqa: B950

    def test_densenet(self) -> None:
        """Tests densenet* retrieval."""
        model = _get_torchvision_classification_model("densenet169", False, 2)
        assert model.classifier.out_features == 2  # type: ignore[union-attr]  # Reason: test will fail if wrong type  # noqa: B950

    def test_squeezenet(self) -> None:
        """Tests squeezenet* retrieval."""
        model = _get_torchvision_classification_model("squeezenet1_0", False, 2)
        assert model.classifier[1].out_channels == 2  # type: ignore[index,union-attr]  # Reason: test will fail if wrong type  # noqa: B950
        assert model.num_classes == 2



"""Tests for MessageService federated transport layer with PyTorch models."""
import asyncio
import os
from typing import Callable
from unittest.mock import Mock

import pytest
from pytest import fixture

from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
from tests.utils.helper import (
    AUC_THRESHOLD,
    backend_test,
    create_local_modeller_and_workers,
    integration_test,
)


@backend_test
@integration_test
class TestMessageServiceUsage:
    """Tests for the message service classes using PyTorch models."""

    @fixture
    def mock_get_pod_public_keys(
        self, apply_mock_get_pod_public_keys: Callable[[str], Mock]
    ) -> Mock:
        """Mocks out get_pod_public_keys function in modeller.py."""
        return apply_mock_get_pod_public_keys(
            "bitfount.federated.modeller._get_pod_public_keys"
        )

    async def test_classification_results_only(
        self,
        mock_get_pod_public_keys: Mock,
        mock_message_aes_decryption: Mock,
        mock_message_aes_encryption: Mock,
        mock_rsa_encryption: Mock,
        mock_rsa_sign_message: Mock,
    ) -> None:
        """Tests PyTorchTabularClassifier with ResultsOnly protocol."""
        modeller, workers = create_local_modeller_and_workers(
            model_name="PyTorchTabularClassifier",
            protocol_name="ResultsOnly",
            algorithm_name="ModelTrainingAndEvaluation",
        )
        modeller_results, *_ = await asyncio.gather(
            modeller.run_async(
                [worker.mailbox.pod_identifier for worker in workers], True
            ),
            *[_run_func_and_listen_to_mailbox(w.run(), w.mailbox) for w in workers]
        )

        assert isinstance(modeller_results, dict)
        for result in modeller_results.values():
            assert result is not None
            assert isinstance(result, dict)
            auc = result["AUC"]
            assert auc > AUC_THRESHOLD

    async def test_classification_federated_averaging_and_early_stopping_autoeval_true(
        self,
        mock_get_pod_public_keys: Mock,
        mock_message_aes_decryption: Mock,
        mock_message_aes_encryption: Mock,
        mock_rsa_encryption: Mock,
        mock_rsa_sign_message: Mock,
    ) -> None:
        """Tests PyTorchTabularClassifier with FederatedAveraging and EarlyStopping."""
        modeller, workers = create_local_modeller_and_workers(
            model_name="PyTorchTabularClassifier",
            protocol_name="FederatedAveraging",
            algorithm_name="FederatedModelTraining",
            early_stopping=True,
        )

        modeller_results, *_ = await asyncio.gather(
            modeller.run_async(
                [worker.mailbox.pod_identifier for worker in workers], True
            ),
            *[_run_func_and_listen_to_mailbox(w.run(), w.mailbox) for w in workers]
        )

        assert modeller_results is not None
        auc = float(modeller_results[0]["AUC"])
        assert auc > AUC_THRESHOLD

    async def test_classification_federated_averaging_and_early_stopping_autoeval_false(
        self,
        mock_get_pod_public_keys: Mock,
        mock_message_aes_decryption: Mock,
        mock_message_aes_encryption: Mock,
        mock_rsa_encryption: Mock,
        mock_rsa_sign_message: Mock,
    ) -> None:
        """Tests PyTorchTabularClassifier with FederatedAveraging and EarlyStopping."""
        modeller, workers = create_local_modeller_and_workers(
            model_name="PyTorchTabularClassifier",
            protocol_name="FederatedAveraging",
            algorithm_name="FederatedModelTraining",
            early_stopping=True,
            auto_eval=False,
        )

        modeller_results, *_ = await asyncio.gather(
            modeller.run_async(
                [worker.mailbox.pod_identifier for worker in workers], True
            ),
            *[_run_func_and_listen_to_mailbox(w.run(), w.mailbox) for w in workers]
        )
        assert modeller_results == []

    @pytest.mark.skipif(
        bool(os.getenv("GITHUB_ACTIONS", False)),
        reason=(
            "Multi-runner tests pass locally but not on constrained GHA runner"
            " systems. Will be addressed in BIT-2092."
        ),
    )
    async def test_classification_secure_aggregation(
        self,
        mock_get_pod_public_keys: Mock,
        mock_message_aes_decryption: Mock,
        mock_message_aes_encryption: Mock,
        mock_rsa_decryption: Mock,
        mock_rsa_encryption: Mock,
        mock_rsa_sign_message: Mock,
    ) -> None:
        """Tests PyTorchTabularClassifier Federated Averaging and Secure Aggregation."""
        modeller, workers = create_local_modeller_and_workers(
            model_name="PyTorchTabularClassifier",
            protocol_name="FederatedAveraging",
            algorithm_name="FederatedModelTraining",
            secure_aggregation=True,
        )
        modeller_results, *_ = await asyncio.gather(
            modeller.run_async(
                [worker.mailbox.pod_identifier for worker in workers], True
            ),
            *[_run_func_and_listen_to_mailbox(w.run(), w.mailbox) for w in workers]
        )

        assert modeller_results is not None
        auc = float(modeller_results[0]["AUC"])
        assert auc > AUC_THRESHOLD



"""Tests datafactory.py."""
from unittest.mock import create_autospec

from bitfount.backends.pytorch.data.datafactory import _PyTorchDataFactory
from bitfount.backends.pytorch.data.dataloaders import PyTorchIterableBitfountDataLoader
from bitfount.backends.pytorch.data.datasets import _PyTorchIterableDataset
from tests.utils.helper import backend_test, unit_test


@backend_test
@unit_test
def test_datafactory_iterable_dataset() -> None:
    """Tests create_dataloader returns iterable loader."""
    datafactory = _PyTorchDataFactory()
    dataset = create_autospec(_PyTorchIterableDataset)
    dataloader = datafactory.create_dataloader(dataset)
    assert isinstance(dataloader, PyTorchIterableBitfountDataLoader)



"""Unary (one reference argument) transformations.

This module contains the base class and concrete classes for unary transformations,
those that take a single reference argument (i.e. a column or transformation name).
"""
from __future__ import annotations

import copy
from functools import cached_property
import logging
from typing import Any, Dict, List, Optional, Union, cast

import attr
from marshmallow import fields, post_load
import marshmallow_union

from bitfount.transformations.base_transformation import (
    MultiColumnOutputTransformation,
    Transformation,
    _TransformationSchema,
)
from bitfount.transformations.exceptions import IncorrectReferenceError
from bitfount.transformations.references import _extract_ref
from bitfount.types import _JSONDict
from bitfount.utils import delegates


@delegates()
@attr.dataclass(kw_only=True)
class UnaryOperation(Transformation):
    """The base abstract class for all Unary Operation Transformations.

    Args:
        arg: The argument to the transformation.
    """

    arg: Any


@delegates()
@attr.dataclass(kw_only=True)
class StringUnaryOperation(UnaryOperation):
    """This class represents any UnaryOperation where arg can only be a string.

    Args:
        arg: The argument to the transformation as a string.
    """

    arg: str


@delegates()
@attr.dataclass(kw_only=True)
class OneHotEncodingTransformation(
    StringUnaryOperation, MultiColumnOutputTransformation
):
    """One hot encoding transformation.

    Represents the transformation of a column into a series of one-hot encoded
    columns.

    Args:
        arg: Column or transformation reference to one-hot encode.
        values: Column values that should be one-hot encoded. This can either be
            a list of values, in which case the one-hot encoding will produce
            columns named `{name}_{value}`, or a dictionary of values to desired
            column suffixes, in which case the encoding will use those suffixes
            (if an entry in the dictionary maps to None, the column name will be
            generated in the same way as described above).

            If `name` is not set, the column or transformation reference from
            `arg` will be used instead.

            Any value found in the column which is not enumerated in this argument
            will be encoded in an `{name}_{unknown_suffix}` column. This column is
            therefore protected and any value or value-column mapping that could
            clash will raise ValueError. If you need to encode such a value,
            `unknown_suffix` must be changed
        unknown_suffix: The suffix to use to create a column for encoding unknown
            values. The column will be created as `{name}_{unknown_suffix}`.
            Default is "UNKNOWN".

    Raises:
        ValueError: If any name in `values` would cause a clash with the unknown
            value column created by `unknown_suffix` or with another generated column.
        ValueError: If no `values` were provided.
        ValueError: If no name is provided and the reference in arg cannot be found.
    """

    _registry_name = "onehotencode"

    values: Dict[Any, str] = attr.ib(init=False, default=None)
    unknown_suffix: str = "UNKNOWN"

    # This is used as a temporary placeholder for the values
    _raw_values: Union[List[Any], Dict[Any, Optional[str]]]
    # This is used to mark if name was set or not
    _no_name_provided: bool = attr.ib(init=False, default=False)

    def __attrs_post_init__(self) -> None:
        # Set this first as the super() call will randomly init name if not set
        if not self.name:
            self._no_name_provided = True
        super().__attrs_post_init__()

        # Check at least one value provided
        if len(self._raw_values) == 0:
            raise ValueError("At least one value must be provided to one-hot encode.")

        self._produce_full_col_map()

    @property
    def columns(self) -> List[str]:
        """Lists the columns that will be output."""
        return list(self.values.values()) + [self.unknown_col]

    @cached_property
    def prefix(self) -> str:
        """Uses name as prefix or extract from arg (should be col or transform ref)."""
        # If name has been randomly set, try to use arg reference instead.
        # If arg isn't a reference, we have to fail out.
        if self._no_name_provided:
            try:
                return _extract_ref(self.arg)
            except IncorrectReferenceError as ire:
                raise ValueError(
                    "No name provided and no reference found in arg."
                ) from ire
        # If name is provided, use it
        else:
            return cast(str, self.name)

    @property
    def unknown_col(self) -> str:
        """Returns the name of the column that unknown values are encoded to."""
        return f"{self.prefix}_{self.unknown_suffix}"

    def _produce_full_col_map(self) -> None:
        """Takes loaded values and produces full value-column map.

        Validates that the values will give unique column names.
        """
        # If self.values already exists, don't redo
        if cast(Optional[Dict[Any, str]], self.values) is not None:
            logging.warning("_produce_full_col_map should not be called more than once")
            return

        prefix = self.prefix

        # Convert list to dict format (provide null mappings for now)
        tmp_values_w_null: Dict[Any, Optional[str]]
        if isinstance(self._raw_values, list):
            tmp_values_w_null = {i: None for i in self._raw_values}
            # Check for duplicate values at this point
            if len(self._raw_values) != len(tmp_values_w_null):
                raise ValueError("If `raw_values` is a list, elements must be unique.")
        else:  # if _raw_values is a dict, copy
            tmp_values_w_null = copy.deepcopy(self._raw_values)

        # Generate column names
        for ohe_val, col_suffix in tmp_values_w_null.items():
            # Use suffix if provided
            if col_suffix:
                col_name = f"{prefix}_{col_suffix}"
            # Otherwise use ohe_val
            else:
                col_name = f"{prefix}_{str(ohe_val)}"
            tmp_values_w_null[ohe_val] = col_name

        # All entries have str values at this point
        tmp_values: Dict[Any, str] = cast(Dict[Any, str], tmp_values_w_null)

        # Check column names are unique
        col_names = tmp_values.values()
        if len(col_names) != len(set(col_names)):
            raise ValueError(
                f"Column names generated must be unique: {sorted(col_names)}"
            )

        # Check each potential column name against the unknown column
        unknown_col = self.unknown_col
        if any(unknown_col == col_name for col_name in col_names):
            raise ValueError(
                f"At least one column name clashes with the unknown value "
                f"column: {unknown_col}"
            )

        # Set self.values to calculated version
        self.values = tmp_values

    class _Schema(_TransformationSchema):
        # From Transformation
        name = fields.String(default=None)
        output = fields.Boolean(default=False)
        # From StringUnaryOperation
        arg = fields.String(required=True)
        # From OneHotEncodingTransformation
        values = marshmallow_union.Union(
            [
                fields.List(fields.Raw),
                fields.Dict(keys=fields.Raw, values=fields.String(allow_none=True)),
            ],
            required=True,
        )
        unknown_suffix = fields.String(default="UNKNOWN")

        @post_load
        def make_transformation(
            self, data: _JSONDict, **_kwargs: Any
        ) -> OneHotEncodingTransformation:
            """Makes a OneHotEncodingTransformation from the schema."""
            # Need to move schema "values" to "raw_values" for __init__ call.
            # It's not "_raw_values" because of how attrs handles private variables.
            raw_values = data.pop("values")
            data["raw_values"] = raw_values
            return OneHotEncodingTransformation(**data)


@delegates()
@attr.dataclass(kw_only=True)
class InclusionTransformation(StringUnaryOperation):
    """Represents the test for substring inclusion in a column's entries.

    Check whether `in_str` (the test string) is in the elements of `arg` (the column).

    Args:
        in_str: The string to test for inclusion.
    """

    _registry_name = "in"

    in_str: str



"""Contains PyTorch specific utility methods."""
from enum import Enum
import logging

import pytorch_lightning.loggers as pl_loggers
import torch

logger = logging.getLogger(__name__)


def _autodetect_gpu() -> int:
    """Detects and returns the number of GPUs available."""
    # Run on GPU if available
    gpus: int = torch.cuda.device_count()
    if gpus > 0:
        gpu_0_name: str = torch.cuda.get_device_name(0)

    # Reduce to 1 GPU if multiple detected
    # TODO: [BIT-492] Add multi-GPU support.
    if gpus > 1:
        logger.warning(
            f"Bitfount model currently only supports one GPU. "
            f"Will use GPU 0 ({gpu_0_name})."
        )
        gpus = 1

    # Log GPU state
    if gpus == 1:
        logger.info(f"CUDA support detected. GPU ({gpu_0_name}) will be used.")
    else:
        logger.info("No supported GPU detected. Running model on CPU.")
    return gpus


class LoggerType(Enum):
    """Different types of loggers for PyTorchLightning.

    With the exception of CSVLogger and TensorBoardLogger, all loggers need to have
    their corresponding python libraries installed separately.

    More information about PyTorchLightning loggers can be found here:
    https://pytorch-lightning.readthedocs.io/en/latest/common/loggers.html
    """

    CSVLogger = pl_loggers.CSVLogger
    MLFlow = pl_loggers.MLFlowLogger
    Neptune = pl_loggers.NeptuneLogger
    TensorBoard = pl_loggers.TensorBoardLogger
    TestTube = pl_loggers.TestTubeLogger
    WeightsAndBiases = pl_loggers.WandbLogger



"""Contains mixin for role-finding code."""
from __future__ import annotations

from enum import Enum, unique
from functools import cached_property
import inspect
from typing import Any, Set, Union


@unique
class Role(Enum):
    """Enum representing roles available in factory classes.

    This applies to Protocols, Algorithms and Aggregators.
    """

    # The value should be the function name that returns
    # classes related to that role.
    MODELLER = "modeller"
    WORKER = "worker"


class _RolesMixIn:
    """A mixin providing a list of roles available in a class."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()

    # Ideally we will make this a property of the class, so an instance is not
    # required to get this information. However, class properties are not supported
    # until Python 3.9 (when @classmethod can wrap @property) and so this is the
    # best solution until then.
    @cached_property
    def roles(self) -> Set[Role]:
        """The set of roles available for this class."""
        available_roles = set()

        # Find roles that have the correct factory methods on this class.
        for role in Role:
            func_name: str = role.value
            # AttributeError will be thrown if the target attribute doesn't exist
            # at all so we catch that during iteration.
            try:
                if inspect.ismethod(getattr(self, func_name)):
                    available_roles.add(role)
            except AttributeError:
                pass
        return available_roles

    def create(self, role: Union[str, Role], **kwargs: Any) -> Any:
        """Create an instance representing the role specified."""
        r_role: Role = Role(role)

        # If role is available for this class, use the underlying factory method.
        if r_role in self.roles:
            fact_func = getattr(self, r_role.value)
            return fact_func(**kwargs)
        # Otherwise, report not supported role
        else:
            raise ValueError(
                f"Role not supported: {r_role} not usable with {type(self)}"
            )



"""Base mailbox classes for other mailbox classes to inherit from."""
import asyncio
from asyncio import FIRST_COMPLETED, AbstractEventLoop, Task
from asyncio.futures import Future as AsyncFuture
from collections import defaultdict
from concurrent.futures import Future as ConcurrentFuture
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager, contextmanager
import inspect
import threading
import time
from typing import (
    Any,
    AsyncContextManager,
    AsyncGenerator,
    Awaitable,
    Callable,
    Coroutine,
    DefaultDict,
    Dict,
    Final,
    Generator,
    Iterable,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)
import warnings

from grpc import RpcError, StatusCode
import msgpack

from bitfount.exceptions import BitfountError
from bitfount.federated.exceptions import (
    MessageHandlerNotFoundError,
    MessageRetrievalError,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.transport.handlers import _PriorityHandler, _TemporaryHandler
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
    _MessageEncryption,
    _MessageService,
    msgpackext_decode,
    msgpackext_encode,
)
from bitfount.federated.transport.utils import _synchronised

logger = _get_federated_logger(__name__)

SyncHandler = Callable[[_BitfountMessage], None]
AsyncHandler = Callable[[_BitfountMessage], Awaitable[None]]
NonPriorityHandler = Union[SyncHandler, AsyncHandler]
Handler = Union[NonPriorityHandler, _PriorityHandler]

HANDLER_BACKOFF_SECONDS: int = 30

# Additional types for handler registration storage
ANY_MESSAGE: Final = "ANY_MESSAGE"
_ExtendedMessageTypes = Union[_BitfountMessageType, Literal["ANY_MESSAGE"]]
# Registered handlers can additionally be of type _TemporaryHandler
_RegisterHandler = Union[Handler, _TemporaryHandler]
_HandlersDict = Dict[_RegisterHandler, Optional[asyncio.Lock]]


@asynccontextmanager
async def _asyncnullcontext() -> AsyncGenerator[None, None]:
    """Async version of contextlib.nullcontext()."""
    yield None


@asynccontextmanager
async def _async_lock(lock: threading.Lock) -> AsyncGenerator[None, None]:
    """Acquire a threading Lock without blocking the async event loop."""
    loop = asyncio.get_running_loop()
    await loop.run_in_executor(None, lock.acquire)
    try:
        yield  # the lock is held
    finally:
        lock.release()


class _HandlerLockPair(NamedTuple):
    """Pair of handler and associated exclusivity lock."""

    handler: _RegisterHandler
    lock: Optional[asyncio.Lock]


class _HandlerRegistry:
    """Registry for message handlers.

    Thread-safe.
    """

    _registry: DefaultDict[_ExtendedMessageTypes, _HandlersDict]
    _sync_lock: threading.RLock

    def __init__(
        self,
        handlers: Optional[
            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
        ] = None,
    ) -> None:
        self._registry: DefaultDict[_ExtendedMessageTypes, _HandlersDict] = defaultdict(
            dict
        )
        if handlers:
            # Register supplied handlers
            for message_type, handlers_ in handlers.items():
                try:
                    # Assume iterable
                    for handler in cast(Iterable[Handler], handlers_):
                        # TODO: [BIT-1048] Revisit this and decide if _all_ messages
                        #       need to be exclusive in this way.
                        self.register_handler(message_type, handler, exclusive=True)
                except TypeError:
                    # Otherwise, register individual handler
                    self.register_handler(
                        message_type, cast(Handler, handlers_), exclusive=True
                    )

        # Lock for managing multithread operations
        self._sync_lock = threading.RLock()

    @_synchronised
    def register_handler(
        self,
        message_type: _ExtendedMessageTypes,
        handler: _RegisterHandler,
        exclusive: bool = True,
    ) -> None:
        """Registers a handler for a specific message type.

        Args:
            message_type: The message type to register the handler for.
            handler: The handler.
            exclusive: Whether only a single instance of the handler can be running
                at a given time.
        """
        if exclusive:
            if isinstance(handler, _PriorityHandler):
                # The exclusivity locking is handled within the priority handler
                # instead
                lock = None
                handler.set_exclusive()
            else:
                lock = asyncio.Lock()
        else:
            lock = None

        self._registry[message_type][handler] = lock

    @_synchronised
    def delete_handler(
        self,
        message_type: _ExtendedMessageTypes,
        handler: _RegisterHandler,
    ) -> None:
        """Deletes a handler associated with the message type."""
        self._registry[message_type].pop(handler, None)  # avoids KeyError

    @_synchronised
    def delete_all_handlers(self, message_type: _ExtendedMessageTypes) -> None:
        """Deletes all handlers for a specific message type."""
        self._registry[message_type].clear()

    @overload
    def get_handlers(
        self, message_type: _BitfountMessageType, with_locks: Literal[True]
    ) -> List[_HandlerLockPair]:
        """Gets all handlers for a specific message type."""
        ...

    @overload
    def get_handlers(
        self, message_type: _ExtendedMessageTypes, with_locks: Literal[False] = ...
    ) -> List[_RegisterHandler]:
        """Gets all handlers for a specific message type."""
        ...

    @_synchronised  # type: ignore[misc] # Reason: https://github.com/python/mypy/issues/12716 # noqa: B950
    def get_handlers(
        self, message_type: _ExtendedMessageTypes, with_locks: bool = False
    ) -> Union[List[_RegisterHandler], List[_HandlerLockPair]]:
        """Gets all handlers for a specific message type.

        If `with_locks` is True, returns a list of tuples of the handlers and their
        exclusivity locks if present.

        :::caution

        This does not retrieve the universal handlers. If you want to retrieve those
        as well you must call get_handlers(ANY_MESSAGE) and merge the two handler
        lists together.

        :::

        Args:
            message_type: The message type to retrieve the handlers for.
            with_locks: Whether to include handlers' exclusivity locks.

        Returns:
            Either a list of handlers or a list of handler-lock tuples.
        """
        handlers: Union[List[_RegisterHandler], List[_HandlerLockPair]]
        if not with_locks:
            handlers = list(self._registry[message_type].keys())
        else:
            handlers = [
                _HandlerLockPair(k, v) for k, v in self._registry[message_type].items()
            ]

        return handlers

    @_synchronised
    def get_lock(
        self,
        message_type: _ExtendedMessageTypes,
        handler: _RegisterHandler,
    ) -> Optional[asyncio.Lock]:
        """Returns the exclusivity lock for a given handler and message type.

        Returns:
            The exclusivity lock for the handler or None if no lock.

        Raises:
            KeyError: If handler is not associated with message type.
        """
        return self._registry[message_type][handler]


class _BaseMailbox:
    """The base mailbox class.

    Contains handlers and message service.

    Args:
        mailbox_id: the ID of the mailbox to monitor.
        message_service: underlying message service instance.
        handlers: an optional mapping of message types to handlers to initialise with.
    """

    def __init__(
        self,
        mailbox_id: str,
        message_service: _MessageService,
        handlers: Optional[
            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
        ] = None,
    ):
        self.mailbox_id = mailbox_id
        self.message_service = message_service

        self._handlers: _HandlerRegistry = _HandlerRegistry(handlers)

        # Only one call to _listen_for_messages() should be allowed at a time.
        # Otherwise we run the risk of messages being pulled off of the mailbox
        # by a listener that doesn't have the right handlers. Each mailbox should
        # only need one listener as it runs indefinitely.
        self._listening_lock = threading.Lock()

        # To enable a smart back-off when no handler is found before reattempting
        # we introduce an event that can monitor when new handlers are added. This
        # allows us to handle the situation where a response comes through faster
        # than the correct handler can be attached.
        self._new_handler_added = threading.Event()

    async def log(self, message: Mapping[str, object]) -> None:
        """Log message to remote task participant."""
        raise NotImplementedError

    def _setup_federated_logging(self) -> None:
        """Sets up federated logging."""
        raise NotImplementedError

    @contextmanager
    def listen(
        self, handler_dispatch_loop: Optional[AbstractEventLoop] = None
    ) -> Generator[threading.Thread, None, None]:
        """Starts the mailbox listening for messages in a separate thread.

        High-priority messages will be dispatched to a thread pool to be executed
        immediately. Low-priority messages will be dispatched to the event loop
        supplied to this method.

        Args:
            handler_dispatch_loop: The event loop to dispatch low-priority message
                handling to. If not supplied the running event loop from the calling
                thread is used.

        Yields:
            The thread that the mailbox listener is running in.

        Raises:
            BitfountError: If no handler_dispatch_loop is supplied and no event
                loop is running.
        """
        if not handler_dispatch_loop:
            try:
                handler_dispatch_loop = asyncio.get_running_loop()
            except RuntimeError as ex:
                raise BitfountError(
                    "Attempted to run mailbox listener without a running event loop."
                ) from ex

        # Create stop event, so we can notify the underlying thread to cease
        stop_event = threading.Event()

        mailbox_thread = _ThreadWithException(
            target=self._start_in_own_thread,
            name=f"mailbox_thread_{self.mailbox_id}",
            args=(handler_dispatch_loop, stop_event),
            daemon=True,
        )
        mailbox_thread.start()
        try:
            yield mailbox_thread
        finally:
            logger.info(f"Stopping mailbox listener ({self.mailbox_id})...")
            stop_event.set()

            if mailbox_thread.is_alive():
                stop_timeout = 15
                logger.info(
                    f"Waiting up to {stop_timeout} seconds"
                    f" for mailbox listener ({self.mailbox_id}) to stop..."
                )

                mailbox_thread.join(timeout=stop_timeout)

                if mailbox_thread.is_alive():
                    raise TimeoutError(
                        f"Mailbox listener ({self.mailbox_id}) did not stop in time."
                    )

            logger.info(f"Mailbox listener ({self.mailbox_id}) stopped.")

    def _start_in_own_thread(
        self, handler_dispatch_loop: AbstractEventLoop, stop_event: threading.Event
    ) -> None:
        """Start listening for messages in new event loop.

        This is designed to be a thread target.
        """
        return asyncio.run(self._listen_for_messages(handler_dispatch_loop, stop_event))

    async def listen_indefinitely(
        self, handler_dispatch_loop: Optional[AbstractEventLoop] = None
    ) -> None:
        """Listen to a mailbox indefinitely without blocking the event loop.

        Does this by setting up a mailbox listener and dispatcher in a separate
        thread and then providing an async Future to await on. If the listener finishes
        or the caller of this function is cancelled then this will gracefully exit out,
        otherwise it allows the listener to run indefinitely.

        Args:
            handler_dispatch_loop: The event loop to dispatch low-priority message
                handling to. If not supplied the running event loop from the calling
                thread is used.

        Raises:
            BitfountError: If no handler_dispatch_loop is supplied and no event
                loop is running.
        """
        with self.listen(handler_dispatch_loop) as listen_thread:
            # We can't use the ThreadPoolExecutor context manager here because
            # of how the target thread functions and our intentions.
            #
            # If the process is cancelled then we want this function to exit out
            # which will cause the listen_thread to be gracefully shutdown as we
            # leave the self.listen() context manager. If we use the ThreadPoolExecutor
            # context manager it calls executor.shutdown(wait=True) which will cause
            # a deadlock as listen_thread.join() (and hence ThreadPoolExecutor)
            # won't finish until we're out of the self.listen() context manager,
            # but the self.listen() context manager won't finish until we leave
            # the ThreadPoolExecutor context manager.
            #
            # Instead, we need to manually call shutdown on the executor so that
            # it won't block, knowing that the thread in the executor will finish
            # soon because of the graceful shutdown mechanism in self.listen().
            executor = ThreadPoolExecutor(1)
            try:
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(executor, listen_thread.join)
            finally:
                executor.shutdown(wait=False)
        return None

    async def _listen_for_messages(
        self,
        handler_dispatch_loop: AbstractEventLoop,
        stop_event: threading.Event,
    ) -> None:
        """Listens for messages on the target mailbox.

        Received messages are passed to the relevant handlers. If no relevant
        handlers are found, it will wait for up to `HANDLER_BACKOFF_SECONDS`
        for one to be registered. This avoids the situation of a response coming
        through faster than a handler can be registered. If not, it is passed
        to the default handler.
        """
        # Guarantee that only one listener is listening at a time.
        async with _async_lock(self._listening_lock):
            try:
                async for message in self.message_service.poll_for_messages(
                    self.mailbox_id,
                    stop_event,
                ):
                    if stop_event.is_set():
                        logger.info("Stopping listening for messages.")
                        break

                    try:
                        logger.debug(
                            f"Attempting to dispatch handlers for"
                            f" {message.message_type} from {message.sender}"
                        )
                        await self._handle_message(message, handler_dispatch_loop)
                    except MessageHandlerNotFoundError:
                        self._default_handler(message)
            # General message service issues to log out before failing.
            except RpcError as err:
                if err.code() == StatusCode.UNAVAILABLE:
                    logger.warning("Message Service unavailable")
                if err.code() == StatusCode.UNAUTHENTICATED:
                    # This could be a temporary token expiry issue
                    logger.info(
                        f"Authentication to read from '{self.mailbox_id}' failed"
                    )
                if err.code() == StatusCode.PERMISSION_DENIED:
                    logger.warning(
                        f"You don't own a pod with the mailbox: {self.mailbox_id}. "
                        f"Ensure it exists on Bitfount Hub."
                    )
                if err.code() == StatusCode.FAILED_PRECONDITION:
                    logger.debug(
                        f"No mailbox exists for '{self.mailbox_id}', "
                        f"ensure connect_pod() or send_task_requests() is called first."
                    )
                raise MessageRetrievalError(
                    f"An error occurred when trying to communicate"
                    f" with the messaging service: {err}"
                )

    async def _handle_message(
        self,
        message: _BitfountMessage,
        handler_dispatch_loop: AbstractEventLoop,
    ) -> List[Union[ConcurrentFuture, _PriorityHandler]]:
        """Finds and runs handler(s) for the supplied message.

        The lower-priority handler(s) (whether async or not) are run within an
        asyncio event loop (the handler_dispatch_loop) to avoid blocking
        listen_for_messages().

        High-priority handlers are run directly (although in actuality in a thread)
        to avoid a blocking async task from stopping them being executed.

        Args:
            message: The message to handle.
            handler_dispatch_loop: The event loop to dispatch the handler execution
                to. Generally this will be the event loop in the main thread.

        Returns:
            The created concurrent.futures.Futures in which the handler(s) are
            being run or, for high-priority handlers, the _PriorityHandler which
            the execution is tied to.

        Raises:
            MessageHandlerNotFoundError: If no handler is registered for message
                type and one is not registered within the timeout.
        """
        message_type: _BitfountMessageType = message.message_type
        awaitables: List[Union[ConcurrentFuture, _PriorityHandler]] = []

        # Try to find relevant handler(s) for this message type
        handlers: List[_RegisterHandler] = await self._retrieve_handlers(message_type)

        # We create an async wrapper around the handler call regardless of if it's
        # an async function or not to allow us to run it in the background as a Task.
        # This also allows us to access the asyncio.Locks.
        #
        # We need to use a separate method to return the wrapper (rather than defining
        # it in the for-loop) due to the late-binding of closures in Python.
        def _get_running_handler_wrapper(
            handler: NonPriorityHandler,
        ) -> Callable[[], Coroutine[None, None, None]]:
            async def _running_handler_wrapper() -> None:
                # Only a single handler instance (i.e. a call to a specific handler
                # function) should be running at a given time; this helps avoid
                # conflicting use of shared resources and ensures that tasks we
                # _want_ to block (such as worker running) do so.
                # TODO: [BIT-1048] Revisit this and decide if _all_ messages need to
                #       be exclusive in this way.
                message_lock: Optional[AsyncContextManager]
                try:
                    message_lock = self._handlers.get_lock(message_type, handler)
                except KeyError:
                    message_lock = None

                if message_lock is None:
                    message_lock = _asyncnullcontext()

                async with message_lock:
                    # As this supports both sync and async handlers we need to
                    # process the result (which should be None, but could be a
                    # Coroutine returning None). As such, we comfortably call the
                    # handler and then simply await the result if needed.
                    logger.debug(f"Handling {message.message_type} message")
                    result = handler(message)
                    if inspect.isawaitable(result):
                        # Mypy needs some assurance, despite the above check
                        result = cast(Awaitable[None], result)
                        await result

            return _running_handler_wrapper

        for handler_ in handlers:
            handler_callable: Handler
            if not isinstance(handler_, _TemporaryHandler):
                handler_callable = handler_
            else:
                handler_callable = handler_.handler

            if not isinstance(handler_callable, _PriorityHandler):
                # For non-priority handlers we farm them out to the handler dispatch
                # event loop. We return the concurrent Future instance so the task
                # can be monitored later.
                wrapped_handler = _get_running_handler_wrapper(handler_callable)
                fut = asyncio.run_coroutine_threadsafe(
                    wrapped_handler(), handler_dispatch_loop
                )
                awaitables.append(fut)
            else:
                # The handler register locks aren't used for _PriorityHandler as
                # they get run in other threads and asyncio.Locks aren't thread
                # safe. The _PriorityHandler itself enforces locking.
                handler_callable(message)
                awaitables.append(handler_callable)

            # As handler has been dispatched, we should delete it if it's a
            # temporary handler
            if isinstance(handler_, _TemporaryHandler):
                logger.debug(
                    f"Temporary handler {handler_} has been dispatched."
                    f" Deleting handler."
                )
                self._delete_handler(message_type, handler_)

        return awaitables

    async def _retrieve_handlers(
        self, message_type: _BitfountMessageType, timeout: int = HANDLER_BACKOFF_SECONDS
    ) -> List[_RegisterHandler]:
        """Retrieve the registered handler(s) for the given message type.

        If no handler(s) are registered for message type, wait for up to `timeout`
        for one to be registered. Otherwise, raise KeyError.

        Args:
            message_type: Message type to retrieve handler for.
            timeout: Number of seconds to wait before declaring no handler found.

        Returns:
            The handler(s) registered for the given message type.

        Raises:
            MessageHandlerNotFoundError: If no handler is registered for message
                                         type and one is not registered within
                                         the timeout.
        """
        # If handler(s) are already registered, `_retrieve_handlers_backoff()` will
        # return them immediately. Otherwise, we wait for it to return for at most
        # `timeout` seconds before raising MessageHandlerNotFoundError.
        try:
            # The two timeout uses are different; the second is how long `await`
            # will wait before calling the whole thing off. The first is used within
            # _retrieve_handlers_backoff to limit the amount of time to wait on
            # the threading.Event.
            handlers = await asyncio.wait_for(
                self._retrieve_handlers_backoff(message_type, timeout),
                timeout=timeout,
            )
            return handlers
        except asyncio.TimeoutError as te:
            # This is only for debug purposes as the MessageHandlerNotFoundError
            # will be swallowed by the checks in _listen_for_messages and the default
            # handler will be used to process the message.
            logger.debug(
                f"No handler found within backoff period "
                f"for message type {message_type}."
            )
            raise MessageHandlerNotFoundError from te

    async def _retrieve_handlers_backoff(
        self,
        message_type: _BitfountMessageType,
        timeout: float,
    ) -> List[_RegisterHandler]:
        """Wait for handler(s) to be available for message type and return them."""
        # We give a small amount of buffer to ensure the Event waiting doesn't
        # finish before the asyncio.wait_for call.
        timeout *= 1.1
        start_time = time.time()

        # See if we have a relevant handler(s), try to return immediately
        try:
            # If no handlers registered, will return an empty list
            if handlers := self._handlers.get_handlers(message_type):
                # Extend it with any universal handlers
                handlers.extend(self._handlers.get_handlers(ANY_MESSAGE))

                # Deduplicate before returning
                # TODO: [BIT-2248] Deduplication will not be needed or at least
                #       will be different.
                return list(set(handlers))
            else:
                raise KeyError(
                    f"No handlers registered for message type {message_type}"
                )
        except KeyError:
            # If not, loop indefinitely, waiting on handler registration events
            while True:
                # Get remaining timeout based on start time and current time
                remaining_timeout = timeout - (time.time() - start_time)

                # Wait for a new handler to be registered. If one is already registered
                # since the last time this backoff was attempted, returns immediately.
                # As self._new_handler_added is a `threading.Event` we run the waiting
                # in a separate thread to not block the event loop.
                loop = asyncio.get_running_loop()
                wait_fut = loop.run_in_executor(
                    None, self._new_handler_added.wait, remaining_timeout
                )
                event_set: bool = await wait_fut

                if event_set:
                    # Clear the handler registration event, so we can monitor when
                    # another handler is registered.
                    #
                    # We do this at the end of the iteration so that we avoid a race
                    # condition between when `_retrieve_handlers_backoff()` is called
                    # and when it is run where a handler could be registered in that
                    # gap. Clearing the event at the start of the iteration would mean
                    # that handler was missed.
                    self._new_handler_added.clear()
                    try:
                        # See if the newly registered handler is for the message type
                        # we are interested in.
                        # If no handlers registered, will return an empty list
                        if handlers := self._handlers.get_handlers(message_type):
                            # Extend it with any universal handlers
                            handlers.extend(self._handlers.get_handlers(ANY_MESSAGE))

                            # Deduplicate before returning
                            # TODO: [BIT-2248] Deduplication will not be needed
                            #       or at least will be different.
                            return list(set(handlers))
                        else:
                            raise KeyError(
                                f"No handlers registered"
                                f" for message type {message_type}"
                            )
                    except KeyError:
                        # No handlers currently registered for message type
                        pass

    def _register_handler(
        self, message_type: _ExtendedMessageTypes, handler: _RegisterHandler
    ) -> None:
        """Registers a handler for a specific message type."""
        self._handlers.register_handler(message_type, handler)

        # Note we've added a new handler, to allow backed off handler retrieval
        # to know.
        self._new_handler_added.set()

    def register_handler(
        self,
        message_type: _BitfountMessageType,
        handler: Handler,
        high_priority: bool = False,
    ) -> Handler:
        """Registers a handler for a specific message type.

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered handler, which may not be the same as the supplied handler.
        """
        handler = self._process_handler(handler, high_priority)

        logger.debug(f"Registering handler {handler} for message type {message_type}")
        self._register_handler(message_type, handler)

        return handler

    def register_universal_handler(
        self, handler: Handler, high_priority: bool = False
    ) -> Handler:
        """Registers a universal handler, that will also be called for all messages.

        :::caution

        Universal handlers are run IN ADDITION to any other handlers for that message
        type. If no non-universal handlers are registered for that message type,
        the default handler will be used instead.

        :::

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered handler, which may not be the same as the supplied handler.
        """
        handler = self._process_handler(handler, high_priority)

        logger.debug(f"Registering universal handler {handler}")
        self._register_handler(ANY_MESSAGE, handler)

        return handler

    def register_temp_handler(
        self,
        message_type: _BitfountMessageType,
        handler: Handler,
        high_priority: bool = False,
    ) -> _TemporaryHandler:
        """Registers a handler that will be deleted after it is called.

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered temporary handler.
        """
        handler = self._process_handler(handler, high_priority)
        temp_handler = self._make_temp_handler(handler)

        logger.debug(
            f"Registering temporary handler {handler} for message type {message_type}"
        )
        self._register_handler(message_type, temp_handler)

        return temp_handler

    def register_temp_universal_handler(
        self, handler: Handler, high_priority: bool = False
    ) -> _TemporaryHandler:
        """Registers a universal handler that will be deleted after it is called.

        :::caution

        Universal handlers are run IN ADDITION to any other handlers for that message
        type. If no non-universal handlers are registered for that message type,
        the default handler will be used instead.

        :::

        If `high_priority` is true, the handler will be converted to a high priority
        handler. Note that only synchronous functions are compatible with
        `high_priority`.

        Returns:
            The registered temporary handler.
        """
        handler = self._process_handler(handler, high_priority)
        temp_handler = self._make_temp_handler(handler)

        logger.debug(f"Registering temporary universal handler {handler}")
        self._register_handler(ANY_MESSAGE, temp_handler)

        return temp_handler

    @staticmethod
    def _make_temp_handler(handler: Handler) -> _TemporaryHandler:
        """Wraps handler so that it will be deleted after it is called."""
        return _TemporaryHandler(handler)

    def _delete_handler(
        self, message_type: _BitfountMessageType, handler: _RegisterHandler
    ) -> None:
        """Delete a handler from the registry."""
        self._handlers.delete_handler(message_type, handler)

    def delete_handler(
        self,
        message_type: _BitfountMessageType,
        handler: Optional[_RegisterHandler] = None,
    ) -> None:
        """Deletes a handler associated with the message type.

        If a specific handler is not provided, deletes all handlers for that
        message type.
        """
        if not handler:
            warnings.warn(
                "In future versions, delete_handler will require a specific handler"
                " instance to be provided. Please switch to delete_all_handlers"
                " instead to maintain previous functionality.",
                DeprecationWarning,
            )
            self.delete_all_handlers(message_type)
        else:
            logger.debug(f"Deleting handler for message type: {message_type}")
            self._delete_handler(message_type, handler)

    def delete_universal_handler(self, handler: _RegisterHandler) -> None:
        """Deletes a specific universal handler."""
        logger.debug(f"Deleting universal handler {handler}")
        self._handlers.delete_handler(ANY_MESSAGE, handler)

    def delete_all_handlers(self, message_type: _BitfountMessageType) -> None:
        """Deletes all handlers for a specific message type."""
        logger.debug(f"Deleting all handlers for message type: {message_type}")
        self._handlers.delete_all_handlers(message_type)

    def delete_all_universal_handlers(self) -> None:
        """Deletes all universal handlers."""
        logger.debug("Deleting all universal handlers")
        self._handlers.delete_all_handlers(ANY_MESSAGE)

    @staticmethod
    def _default_handler(message: _BitfountMessage) -> None:
        """Simple default handler that logs the message details.

        If this is called it is because we have received a message type that we
        were not expecting and do not know how to handle. We log out pertinent
        (non-private) details.
        """
        logger.error(
            f"Received unexpected message "
            f"("
            f"type: {message.message_type}; "
            f"sender {message.sender}; "
            f"recipient {message.recipient}"
            f"). "
            f"Message was not handled."
        )

    @staticmethod
    def _process_handler(handler: Handler, high_priority: bool) -> Handler:
        """Process the supplied handler given handler configuration.

        Handles conversion into high priority handlers as needed and also does
        type-checking for supported handler types.

        Return:
            The processed (and hence potentially changed) handler.
        """
        if high_priority:
            if isinstance(handler, _PriorityHandler):
                # Already correct, just return
                return handler
            else:
                logger.debug(f"Converting handler {handler} to high priority handler.")
                return _PriorityHandler(handler)
        else:
            if isinstance(handler, _PriorityHandler):
                logger.warning(
                    "A priority handler has been provided but high_priority not set."
                    " Treating as though high_priority is True."
                )
                return handler
            else:
                return handler


async def _send_aes_encrypted_message(
    message: Any,
    aes_encryption_key: bytes,
    message_service: _MessageService,
    **kwargs: Any,
) -> None:
    """Packs message, encrypts it and sends it.

    Args:
        message: The message to be sent. Must support serialisation via msgpack.
        aes_encryption_key: Key used to encrypt message.
        message_service: The MessageService used to send the message.
        **kwargs: Keyword arguments passed to BitfountMessage constructor.
    """
    body = msgpack.dumps(message, default=msgpackext_encode)
    message_body = _MessageEncryption.encrypt_outgoing_message(body, aes_encryption_key)

    await message_service.send_message(
        _BitfountMessage(body=message_body, **kwargs),
        already_packed=True,
    )


def _decrypt_aes_message(message: _BitfountMessage, aes_encryption_key: bytes) -> Any:
    """Decrypt AES encrypted message.

    Args:
        message: Encrypted message to decrypt.
        aes_encryption_key: The AES key to use to decrypt.

    Returns:
        Decrypted message body.
    """
    body = _MessageEncryption.decrypt_incoming_message(message.body, aes_encryption_key)
    return msgpack.loads(body, ext_hook=msgpackext_decode)


class _ThreadWithException(threading.Thread):
    def run(self) -> None:
        self._exc: Optional[Exception] = None
        try:
            super().run()
        except Exception as e:
            self._exc = e
            raise e

    def join(self, timeout: Optional[float] = None) -> None:
        super().join(timeout)
        if self._exc:
            raise self._exc


# Type variable for return type below.
_R = TypeVar("_R")


async def _run_func_and_listen_to_mailbox(
    run_func: Coroutine[None, None, _R],
    mailbox: _BaseMailbox,
) -> _R:
    """Runs an async function and listens for messages simultaneously.

    This function allows any exceptions that occur in the run function to be properly
    propagated to the calling code whilst ensuring that the listener or run function
    are correctly shutdown in such a situation.

    It also ensures that the mailbox listener is not run for longer than the lifetime
    of run_func.

    Args:
        run_func: The function to run that will be needing received messages.
        mailbox: The mailbox to use to listen for messages.

    Returns:
         The return value of run_func.
    """
    with mailbox.listen() as mailbox_listener:
        loop = asyncio.get_running_loop()
        executor = ThreadPoolExecutor(1)
        try:
            # Create a separate task for the wrapped function call
            run_task: Task[_R] = asyncio.create_task(run_func)

            # Create future to monitor mailbox listener
            fut: AsyncFuture[None] = loop.run_in_executor(
                executor, mailbox_listener.join
            )

            aws: List[Awaitable] = [run_task, fut]
            done, pending = await asyncio.wait(aws, return_when=FIRST_COMPLETED)

            if run_task in done:
                # Task has completed, return the result (or raise an exception)
                logger.debug(f"Task {run_task} is done")
                return run_task.result()
            else:  # run_task not in done
                # The mailbox listener has finished prematurely (likely due to an
                # error).
                exc = fut.exception()
                error_msg = (
                    f"Mailbox has finished listening before the target task"
                    f" ({run_task}) finished running."
                )

                if exc:
                    # If an exception occurred in the mailbox listener, log information
                    # about it and then throw the exception into the run_task to handle
                    error_msg += (
                        f" Exception was: {exc}. This has been passed to the task"
                        f" to handle."
                    )
                    logger.error(error_msg)
                    logger.exception(exc)

                    # Pass to task to handle then await on task
                    cast(Coroutine, run_task.get_coro()).throw(exc)
                    return await run_task
                else:
                    # Otherwise, log that something went wrong and wait on the run_task
                    # to "finish"
                    logger.error(error_msg)
                    return await run_task
        finally:
            executor.shutdown(wait=False)



"""Tests for the private set intersection protocol."""
import re
from typing import Callable
from unittest.mock import Mock, create_autospec

import pytest
from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.compute_intersection_rsa import (
    ComputeIntersectionRSA,
)
from bitfount.federated.exceptions import PSINoDataSourceError
from bitfount.federated.protocols import psi
from bitfount.federated.protocols.base import (
    _BaseModellerProtocol,
    _BaseProtocol,
    _BaseWorkerProtocol,
    registry,
)
from bitfount.federated.protocols.psi import (
    PrivateSetIntersection,
    _ModellerSide,
    _PSICompatibleAlgoFactory_,
    _PSICompatibleModeller,
    _PSICompatibleWorker,
    _WorkerSide,
)
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils.helper import create_datasource, unit_test


@unit_test
class TestWorkerSide:
    """Tests for the worker-side of the PrivateSetIntersection."""

    @fixture
    def mock_datasource(self) -> Mock:
        """Mock BaseSource."""
        mock_datasource: Mock = create_autospec(BaseSource, instance=True)
        return mock_datasource

    @fixture
    def mock_mailbox(self) -> Mock:
        """Mock WorkerMailbox."""
        mock_mailbox: Mock = create_autospec(_WorkerMailbox, instance=True)
        return mock_mailbox

    @fixture
    def worker_side_factory(self, mock_mailbox: Mock) -> Callable[[Mock], _WorkerSide]:
        """Factory to create WorkerSide instances from mock algorithms."""

        def _create(algo: Mock) -> _WorkerSide:
            return _WorkerSide(
                algorithm=algo,
                mailbox=mock_mailbox,
            )

        return _create

    async def test_run(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Tests WorkerSide.run() for PrivateSetIntersection."""
        mock_algorithm: Mock = create_autospec(_PSICompatibleWorker, instance=True)
        mock_algorithm.public_key = Mock()
        worker_side: _WorkerSide = worker_side_factory(mock_algorithm)
        mock_modeller_psi_dataset = mocker.patch.object(
            worker_side, "_receive_modeller_psi_dataset"
        )

        await worker_side.run(mock_datasource)
        mock_modeller_psi_dataset.assert_awaited_once()

        mock_algorithm.initialise.assert_called_once_with(
            datasource=mock_datasource, columns_to_intersect=None, table=None
        )
        mock_algorithm.run.assert_called_once()
        mock_mailbox.get_task_complete_update.assert_called_once()

    async def test__receive_modeller_psi_dataset(
        self,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Tests that _receive_modeller_psi_dataset awaits _get_psi_dataset."""
        mock_algorithm: Mock = create_autospec(_PSICompatibleWorker, instance=True)
        mock_modeller_psi_dataset = mocker.patch(
            "bitfount.federated.protocols.psi._get_psi_dataset", return_value=["123"]
        )
        worker_side: _WorkerSide = worker_side_factory(mock_algorithm)
        data = await worker_side._receive_modeller_psi_dataset()
        mock_modeller_psi_dataset.assert_awaited_once_with(mock_mailbox)
        assert data[0] == 123

    async def test__send_public_key(
        self,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Tests that _send_public_key_to_modeller awaits _send_public_key."""
        mock_algorithm: Mock = create_autospec(_PSICompatibleWorker, instance=True)
        mock__send_public_key = mocker.patch(
            "bitfount.federated.protocols.psi._send_public_key"
        )
        worker_side: _WorkerSide = worker_side_factory(mock_algorithm)
        await worker_side._send_public_key_to_modeller(public_key=Mock)  # type: ignore[arg-type] # Reason: mocking for testing purposes # noqa: B950
        mock__send_public_key.assert_awaited_once()

    async def test__send_psi_data_to_modeller(
        self,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        worker_side_factory: Callable[[Mock], _WorkerSide],
    ) -> None:
        """Tests that _send_psi_data_to_modeller awaits _send_psi_dataset_worker."""
        mock_algorithm: Mock = create_autospec(_PSICompatibleWorker, instance=True)
        mock__send_psi_data_to_modeller = mocker.patch(
            "bitfount.federated.protocols.psi._send_psi_dataset_worker"
        )
        worker_side: _WorkerSide = worker_side_factory(mock_algorithm)
        await worker_side._send_psi_data_to_modeller(dataset=([1, 2], [0, 2]))
        mock__send_psi_data_to_modeller.assert_awaited_once_with(
            (["1", "2"], ["0", "2"]), mock_mailbox
        )


@unit_test
class TestModellerSide:
    """Tests for the modeller-side of the PrivateSetIntersection."""

    @fixture
    def mock_datasource(self) -> Mock:
        """Mock BaseSource."""
        mock_datasource: Mock = create_autospec(BaseSource, instance=True)
        return mock_datasource

    @fixture
    def mock_mailbox(self) -> Mock:
        """Mock WorkerMailbox."""
        mock_mailbox: Mock = create_autospec(_ModellerMailbox, instance=True)
        return mock_mailbox

    @fixture
    def modeller_side_factory(
        self, mock_mailbox: Mock
    ) -> Callable[[Mock], _ModellerSide]:
        """Factory to create ModellerSide instances from mock algorithms."""

        def _create(algo: Mock) -> _ModellerSide:
            return _ModellerSide(
                algorithm=algo,
                datasource=Mock(),
                mailbox=mock_mailbox,
            )

        return _create

    async def test_run(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        modeller_side_factory: Callable[[Mock], _ModellerSide],
    ) -> None:
        """Tests ModellerSide.run() for PrivateSetIntersection."""
        mock_algorithm: Mock = create_autospec(_PSICompatibleModeller, instance=True)
        modeller_side: _ModellerSide = modeller_side_factory(mock_algorithm)
        mock__receive_public_key = mocker.patch.object(
            modeller_side, "_receive_public_key"
        )
        mock_get_psi_datasets_from_workers = mocker.patch(
            "bitfount.federated.protocols.psi._get_psi_datasets_from_workers"
        )
        await modeller_side.run()
        mock__receive_public_key.assert_awaited_once()
        mock_get_psi_datasets_from_workers.assert_awaited_once()
        mock_algorithm.initialise.assert_called_once()
        mock_algorithm.run.assert_called_once()

    async def test__send_psi_data_modeller(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        modeller_side_factory: Callable[[Mock], _ModellerSide],
    ) -> None:
        """Tests ModellerSide._send_psi_data_modeller() for PrivateSetIntersection."""
        mock_algorithm: Mock = create_autospec(_PSICompatibleModeller, instance=True)
        mock__send_psi_data_modeller = mocker.patch(
            "bitfount.federated.protocols.psi._send_psi_dataset_modeller"
        )
        modeller_side: _ModellerSide = modeller_side_factory(mock_algorithm)
        await modeller_side._send_psi_data_modeller(dataset=[1, 2])
        mock__send_psi_data_modeller.assert_awaited_once()

    async def test__receive_psi_datasets_from_workers(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        modeller_side_factory: Callable[[Mock], _ModellerSide],
    ) -> None:
        """Tests ModellerSide._receive_psi_datasets_from_workers() for PrivateSetIntersection."""  # noqa: B950
        mock_algorithm: Mock = create_autospec(_PSICompatibleModeller, instance=True)
        mock_get_psi_data = mocker.patch(
            "bitfount.federated.protocols.psi._get_psi_datasets_from_workers"
        )
        modeller_side: _ModellerSide = modeller_side_factory(mock_algorithm)
        await modeller_side._receive_psi_datasets_from_workers()
        mock_get_psi_data.assert_awaited_once()

    async def test__receive_public_key(
        self,
        mock_datasource: Mock,
        mock_mailbox: Mock,
        mocker: MockerFixture,
        modeller_side_factory: Callable[[Mock], _ModellerSide],
    ) -> None:
        """Tests ModellerSide._receive_public_key() for PrivateSetIntersection."""
        mock_algorithm: Mock = create_autospec(_PSICompatibleModeller, instance=True)
        mock__receive_public_key = mocker.patch(
            "bitfount.federated.protocols.psi._get_public_key", return_value=["pub_key"]
        )
        mock_load_public_key = mocker.patch(
            "bitfount.federated.protocols.psi.load_pem_public_key"
        )
        modeller_side: _ModellerSide = modeller_side_factory(mock_algorithm)
        await modeller_side._receive_public_key()
        mock__receive_public_key.assert_awaited_once_with(mock_mailbox)
        mock_load_public_key.assert_called_once()


@unit_test
class TestPrivateSetIntersection:
    """Test PrivateSetIntersection protocol."""

    @fixture(scope="function")
    def remote_algorithm(self) -> ComputeIntersectionRSA:
        """Returns remote algorithm."""
        return ComputeIntersectionRSA()

    def test_modeller(
        self, mock_modeller_mailbox: Mock, remote_algorithm: ComputeIntersectionRSA
    ) -> None:
        """Test modeller method."""
        datasource = create_datasource(classification=True)
        protocol_factory = PrivateSetIntersection(
            algorithm=remote_algorithm, datasource=datasource
        )
        protocol = protocol_factory.modeller(mailbox=mock_modeller_mailbox)
        for type_ in [
            _BaseProtocol,
            _BaseModellerProtocol,
            psi._ModellerSide,
        ]:
            assert isinstance(protocol, type_)

    def test_modeller_no_datasource_error(
        self, mock_modeller_mailbox: Mock, remote_algorithm: ComputeIntersectionRSA
    ) -> None:
        """Test modeller method raises error with no datasource."""
        protocol_factory = PrivateSetIntersection(algorithm=remote_algorithm)
        with pytest.raises(PSINoDataSourceError):
            protocol_factory.modeller(mailbox=mock_modeller_mailbox)

    def test_worker(
        self,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        remote_algorithm: ComputeIntersectionRSA,
    ) -> None:
        """Test modeller method."""
        protocol_factory = PrivateSetIntersection(algorithm=remote_algorithm)
        protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
        for type_ in [
            _BaseProtocol,
            _BaseWorkerProtocol,
            psi._WorkerSide,
        ]:
            assert isinstance(protocol, type_)

    def test__validate_algorithm_accepts(self) -> None:
        """Tests _validate_algorithm accepts compatible."""
        # Test with PSICompatibleAlgoFactory
        mock_algorithm: Mock = create_autospec(
            _PSICompatibleAlgoFactory_, instance=True
        )
        PrivateSetIntersection._validate_algorithm(mock_algorithm)

    def test__validate_algorithm_rejects(self) -> None:
        """Tests _validate_algorithm rejects incompatible."""
        mock_algorithm: Mock = Mock(spec_set=["__name__"])
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"The {PrivateSetIntersection.__name__} protocol does not "
                f"support the {type(mock_algorithm).__name__} algorithm."
            ),
        ):
            PrivateSetIntersection._validate_algorithm(mock_algorithm)

    def test_run_protocol(self, mocker: MockerFixture) -> None:
        """Tests we can run a protocol."""
        algorithm = ComputeIntersectionRSA()
        # Mock out Modeller creation
        mock_modeller = mocker.patch(
            "bitfount.federated.protocols.base._Modeller", autospec=True
        )
        mock_modeller.return_value = mock_modeller  # for __init__
        mock_modeller.run.return_value = None

        protocol = PrivateSetIntersection(algorithm=algorithm)
        protocol.run(
            pod_identifiers=["fake/fake"],
            hub=Mock(),
            message_service=Mock(),
        )

        mock_modeller.run.assert_called_once_with(["fake/fake"], require_all_pods=False)


@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for PSI protocol."""

    def test_serialization(self) -> None:
        """Test Marshmallow Serialization for a ComputeIntersectionRSA algorithm."""
        algorithm_factory = ComputeIntersectionRSA()
        proto = PrivateSetIntersection(algorithm=algorithm_factory)
        dumped = bf_dump(proto)
        loaded = bf_load(dumped, registry)
        assert proto.class_name == loaded.class_name
        from tests.bitfount.backends.pytorch.models.test_models import assert_vars_equal

        assert_vars_equal(vars(proto.algorithm), vars(loaded.algorithm))

    def test_serialization_w_columns_table(self) -> None:
        """Test Marshmallow Serialization for a ComputeIntersectionRSA algorithm."""
        algorithm_factory = ComputeIntersectionRSA()
        proto = PrivateSetIntersection(
            algorithm=algorithm_factory,
            datasource=create_datasource(classification=True),
            datasource_columns=["test"],
            datasource_table="table",
            pod_columns=["test"],
            pod_table="table",
        )
        dumped = bf_dump(proto)
        loaded = bf_load(dumped, registry)
        assert proto.class_name == loaded.class_name
        assert loaded.pod_columns == proto.pod_columns
        assert loaded.pod_table == proto.pod_table
        # Check that the modeller's datasource and columns don't get serialized
        assert loaded._modeller_cols is None
        assert loaded._modeller_table is None
        assert loaded.datasource is None

    def test_dump(self) -> None:
        """Test protocol.dump()."""
        algorithm_factory = ComputeIntersectionRSA()
        proto = PrivateSetIntersection(algorithm=algorithm_factory)
        dumped = proto.dump()
        assert dumped["class_name"] == proto.class_name
        assert dumped["algorithm"]["class_name"] == algorithm_factory.class_name



"""Exceptions related to the transformations package."""
from typing import List, Union

from bitfount.exceptions import BitfountError


class _BitfountTransformationsError(BitfountError):
    """Base class for transformations errors."""


class TransformationRegistryError(_BitfountTransformationsError):
    """Exceptions related to the Transformations registry."""

    pass


class TransformationParsingError(_BitfountTransformationsError):
    """Base exception for transformation parsing."""

    def __init__(self, errors: Union[str, List[str]]):
        """Initialises TransformationParsingError with 1+ error messages.

        Args:
            errors: The parsing error(s) that have occurred.
        """
        if isinstance(errors, str):
            errors = [errors]
        self.errors = errors

    def __str__(self) -> str:
        str_errors = "\n".join(self.errors)
        return f"Errors: \n{str_errors}"


class TransformationProcessorError(_BitfountTransformationsError):
    """Base class for all errors related to transformation processing."""

    def __init__(self, errors: Union[str, List[str]]):
        """Initialises TransformationProcessorError with 1+ error messages.

        Args:
            errors: The parsing error(s) that have occurred.
        """
        if isinstance(errors, str):
            errors = [errors]
        self.errors = errors

    def __str__(self) -> str:
        str_errors = "\n".join(self.errors)
        return f"Errors: \n{str_errors}"


class MissingColumnReferenceError(TransformationProcessorError):
    """Exception for when a column is referenced that doesn't exist."""

    pass


class InvalidBatchTransformationError(TransformationProcessorError):
    """Exception for when a non-batch transformation is attempted with a batch."""

    pass


class TransformationApplicationError(TransformationProcessorError):
    """Exception for when applying the transformation to the data is impossible."""

    pass


class IncorrectReferenceError(_BitfountTransformationsError):
    """Raised when a str is not a column or transformation reference."""

    pass


class NotTransformationReferenceError(IncorrectReferenceError):
    """Raised when a str is not a transformation reference."""

    pass


class NotColumnReferenceError(IncorrectReferenceError):
    """Raised when a str is not a column reference."""

    pass



"""Tests datasplitters.py."""
from typing import Optional, cast
from unittest.mock import MagicMock

from _pytest.logging import LogCaptureFixture
import numpy as np
import pandas as pd
import pytest
from pytest import fixture

from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasplitters import (
    DatasetSplitter,
    PercentageSplitter,
    SplitterDefinedInData,
)
from bitfount.data.types import DataSplit
from tests.utils.helper import DATASET_ROW_COUNT, create_dataset, unit_test


@fixture
def data() -> pd.DataFrame:
    """Returns dataset."""
    return create_dataset()


@unit_test
class TestDataSplitter:
    """Tests abstract DataSplitter."""

    @pytest.mark.parametrize(
        "splitter_name",
        [
            PercentageSplitter.splitter_name(),
            # This helps ensure that it can be mapped to
            # from a string in the config file
            # in case we accidentally change the return type
            # Of the class property above
            "percentage",
        ],
    )
    def test_create_produces_percentage_splitter(self, splitter_name: str) -> None:
        """Test Percentage splitter created."""
        created_splitter = DatasetSplitter.create(splitter_name)
        assert isinstance(created_splitter, PercentageSplitter)

    def test_create_produces_percentage_splitter_with_arguments(self) -> None:
        """Test created percentage splitter has expected values."""
        time_series_sort = ["hello"]

        created_splitter = DatasetSplitter.create(
            PercentageSplitter.splitter_name(),
            validation_percentage=1,
            test_percentage=2,
            time_series_sort_by=time_series_sort,
        )

        assert isinstance(created_splitter, PercentageSplitter)
        assert created_splitter.validation_percentage == 1
        assert created_splitter.test_percentage == 2
        assert created_splitter.time_series_sort_by == time_series_sort

    @pytest.mark.parametrize(
        "splitter_name",
        [
            SplitterDefinedInData.splitter_name(),
            # This helps ensure that it can be mapped to
            # from a string in the config file
            # in case we accidentally change the return type
            # Of the class property above
            "predefined",
        ],
    )
    def test_create_produces_predefined_splitter(self, splitter_name: str) -> None:
        """Test create predefined splitter."""
        created_splitter = DatasetSplitter.create(splitter_name)
        assert isinstance(created_splitter, SplitterDefinedInData)

    def test_create_produces_predefined_splitter_with_arguments(self) -> None:
        """Test created predefined splitter has expected values."""
        column_name = "someColumn"
        training_set_label = "UNIT_TRAIN"
        validation_set_label = "UNIT_VALIDATION"
        test_set_label = "UNIT_TEST"

        created_splitter = DatasetSplitter.create(
            SplitterDefinedInData.splitter_name(),
            column_name=column_name,
            training_set_label=training_set_label,
            validation_set_label=validation_set_label,
            test_set_label=test_set_label,
        )

        assert isinstance(created_splitter, SplitterDefinedInData)
        assert created_splitter.column_name == column_name
        assert created_splitter.training_set_label == training_set_label
        assert created_splitter.validation_set_label == validation_set_label
        assert created_splitter.test_set_label == test_set_label


@unit_test
class TestPercentageSplitter:
    """Tests the Percentage splitter."""

    @pytest.mark.parametrize("validation_percentage", [0, 7])
    @pytest.mark.parametrize("test_percentage", [0, 7])
    def test_sets_split_by_percentage(
        self, data: pd.DataFrame, test_percentage: int, validation_percentage: int
    ) -> None:
        """Test the percentages are as expected."""
        untouched_copy_of_data = data.copy()
        data_splitter = PercentageSplitter(
            validation_percentage=validation_percentage, test_percentage=test_percentage
        )

        (
            train_indices,
            validation_indices,
            test_indices,
        ) = data_splitter.create_dataset_splits(data)

        # Ensure we have the expected number of indices
        assert (
            len(train_indices) + len(validation_indices) + len(test_indices)
            == DATASET_ROW_COUNT
        )
        # Ensure validation and test (and transitively training sets)
        # are the expected sizes
        assert len(validation_indices) == DATASET_ROW_COUNT * (
            validation_percentage / 100
        )
        assert len(test_indices) == DATASET_ROW_COUNT * (test_percentage / 100)

        # Ensure indices are distinct
        # (combines all 3 sets then filters out any duplicates)
        recombined_indices = np.concatenate(
            [train_indices, validation_indices, test_indices]
        )
        unique_indices_returned = np.unique(recombined_indices)
        assert len(unique_indices_returned) == DATASET_ROW_COUNT

        # Ensure we didn't do any sorting
        pd.testing.assert_frame_equal(data, untouched_copy_of_data)

    @pytest.mark.parametrize("validation_percentage", [0, 7])
    @pytest.mark.parametrize("test_percentage", [0, 7])
    def test_time_series(
        self, data: pd.DataFrame, test_percentage: int, validation_percentage: int
    ) -> None:
        """Checks the time series sorting works correctly."""
        data["Date"] = pd.to_datetime(data["Date"], infer_datetime_format=True)
        data = data.reindex(np.random.permutation(data.index.tolist()))

        data_splitter = PercentageSplitter(
            validation_percentage=validation_percentage,
            test_percentage=test_percentage,
            time_series_sort_by="Date",
        )

        (
            train_indices,
            validation_indices,
            test_indices,
        ) = (x.tolist() for x in data_splitter.create_dataset_splits(data))

        # Ensure no overlap of indices between all 3 datasets
        assert not bool(set(train_indices).intersection(validation_indices))
        assert not bool(set(validation_indices).intersection(test_indices))
        assert not bool(set(train_indices).intersection(test_indices))

        train_max = data.loc[train_indices, "Date"].max(skipna=True)
        validation_min = data.loc[validation_indices, "Date"].min(skipna=True)
        validation_max = data.loc[validation_indices, "Date"].max(skipna=True)
        test_min = data.loc[test_indices, "Date"].min(skipna=True)

        # Ensure no overlap of dates between all 3 datasets
        if validation_percentage > 0:
            assert validation_min > train_max

        if test_percentage > 0:
            assert test_min > train_max
            if validation_percentage > 0:
                assert test_min > validation_max

    def test_time_series_multiple_columns(self, data: pd.DataFrame) -> None:
        """Tests time series ordering works when multiple columns are supplied."""
        data["Date"] = pd.to_datetime(data["Date"], infer_datetime_format=True)
        data["year"] = data.Date.dt.year
        data["month"] = data.Date.dt.month

        data = data.reindex(np.random.permutation(data.index.tolist()))
        data_splitter = PercentageSplitter(time_series_sort_by=["year", "month"])

        (
            train_indices,
            validation_indices,
            test_indices,
        ) = (x.tolist() for x in data_splitter.create_dataset_splits(data))

        # Ensure no overlap of indices between all 3 datasets
        assert not bool(set(train_indices).intersection(validation_indices))
        assert not bool(set(validation_indices).intersection(test_indices))
        assert not bool(set(train_indices).intersection(test_indices))

        # Train set
        train = data.loc[train_indices]
        train_max_year, train_max_month = train.loc[train["Date"].idxmax()][
            ["year", "month"]
        ]
        # Validation set
        validation = data.loc[validation_indices]
        validation_min_year, validation_min_month = validation.loc[
            validation["Date"].idxmin()
        ][["year", "month"]]
        validation_max_year, validation_max_month = validation.loc[
            validation["Date"].idxmax()
        ][["year", "month"]]
        # Test set
        test = data.loc[test_indices]
        test_min_year, test_min_month = test.loc[test["Date"].idxmin()][
            ["year", "month"]
        ]

        # Ensure no overlap of dates between all 3 datasets
        assert (
            validation_min_year > train_max_year
            or validation_min_month > train_max_month
        )
        assert (
            test_min_year > validation_max_year or test_min_month > validation_max_month
        )

    @pytest.mark.parametrize(
        "split", [DataSplit.TRAIN, DataSplit.VALIDATION, DataSplit.TEST, None]
    )
    def test_get_split_query(self, split: Optional[DataSplit]) -> None:
        """Tests the get_split_query method."""
        data_splitter = PercentageSplitter()
        mock_datasource = MagicMock(spec=DatabaseSource, query="SELECT * FROM table")
        mock_datasource.__len__.return_value = 1000

        if split is None:
            with pytest.raises(ValueError, match="Split not recognised"):
                data_splitter.get_split_query(mock_datasource, split)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950
        else:
            query = data_splitter.get_split_query(mock_datasource, split)

            if split == DataSplit.TRAIN:
                assert (
                    query == "SELECT * FROM (SELECT * FROM table) q LIMIT 800 OFFSET 0"
                )  # first 80%
            elif split == DataSplit.VALIDATION:
                assert (
                    query
                    == "SELECT * FROM (SELECT * FROM table) q LIMIT 100 OFFSET 800"
                )  # next 10%
            elif split == DataSplit.TEST:
                assert (
                    query
                    == "SELECT * FROM (SELECT * FROM table) q LIMIT 100 OFFSET 900"
                )  # last 10%

    def test_get_split_query_time_series(self, caplog: LogCaptureFixture) -> None:
        """Tests that the get_split_query method logs a warning if time series used."""
        data_splitter = PercentageSplitter(time_series_sort_by="Date")
        mock_datasource = MagicMock(spec=DatabaseSource, query="SELECT * FROM table")
        mock_datasource.__len__.return_value = 1000
        data_splitter.get_split_query(mock_datasource, DataSplit.TRAIN)

        assert (
            "Time series sort by is not supported for Database percentage splits. "
            "The sort by will be ignored. If you want to use time series sort by, "
            "please sort the dataset as you want in the SQL query."
        ) in caplog.text


@unit_test
class TestSplitterDefinedInData:
    """Test SplitterDefinedInData."""

    def test_sets_split(self, data: pd.DataFrame) -> None:
        """Test that the specification in the data is respected."""
        # This is a string column that's present in the dataset
        data_split_column_name = "L"
        training_set_label = "BF_TRAIN"
        validation_set_label = "BF_VALIDATION"
        test_set_label = "BF_TEST"

        data_split_column = np.random.choice(
            [training_set_label, validation_set_label, test_set_label],
            size=data.shape[0],
        )
        data.update(
            # Because name is set, will be coerced into dataframe
            cast(
                pd.DataFrame,
                pd.Series(data_split_column, name=data_split_column_name),
            )
        )

        data = data.reindex(np.random.permutation(data.index.tolist()))
        data_splitter = SplitterDefinedInData(
            column_name=data_split_column_name,
            training_set_label=training_set_label,
            validation_set_label=validation_set_label,
            test_set_label=test_set_label,
        )

        (
            train_indices,
            validation_indices,
            test_indices,
        ) = data_splitter.create_dataset_splits(data)

        # Indices for all rows are included
        assert (
            len(train_indices) + len(validation_indices) + len(test_indices)
            == data.shape[0]
        )

        # We use the indices to get the
        # The training set indices refer to the relevant rows in the dataframe
        assert np.all(data.loc[train_indices].L == training_set_label)
        # The test set indices refer to the relevant rows in the dataframe
        assert np.all(data.loc[test_indices].L == test_set_label)
        # The validation set indices refer to the relevant rows in the dataframe
        assert np.all(data.loc[validation_indices].L == validation_set_label)

    @pytest.mark.parametrize(
        "split", [DataSplit.TRAIN, DataSplit.VALIDATION, DataSplit.TEST, None]
    )
    def test_get_split_query(self, split: Optional[DataSplit]) -> None:
        """Tests the get_split_query method."""
        data_splitter = SplitterDefinedInData()
        mock_datasource = MagicMock(spec=DatabaseSource, query="SELECT * FROM table")
        mock_datasource.__len__.return_value = 1000

        if split is None:
            with pytest.raises(ValueError, match="Split not recognised"):
                data_splitter.get_split_query(mock_datasource, split)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950
        else:
            query = data_splitter.get_split_query(mock_datasource, split)

            if split == DataSplit.TRAIN:
                assert (
                    query == "SELECT * FROM (SELECT * FROM table) q"
                    " WHERE BITFOUNT_SPLIT_CATEGORY = 'TRAIN'"
                )
            elif split == DataSplit.VALIDATION:
                assert (
                    query == "SELECT * FROM (SELECT * FROM table) q"
                    " WHERE BITFOUNT_SPLIT_CATEGORY = 'VALIDATE'"
                )
            elif split == DataSplit.TEST:
                assert (
                    query == "SELECT * FROM (SELECT * FROM table) q"
                    " WHERE BITFOUNT_SPLIT_CATEGORY = 'TEST'"
                )



"""Dataclasses and functionality for task request details/messages."""
from __future__ import annotations

import dataclasses
from dataclasses import dataclass
import typing
from typing import List, Type, TypeVar

import msgpack

from bitfount.federated.types import SerializedProtocol
from bitfount.types import _JSONDict

T = TypeVar("T", bound="_DataclassSerializerMixin")


@dataclass
class _DataclassSerializerMixin:
    """MixIn class for dataclasses that enable easy `msgpack` (de)serialization."""

    def to_dict(self) -> _JSONDict:
        """Returns dataclass as a dictionary."""
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls: Type[T], d: _JSONDict) -> T:
        """Creates dataclass from dictionary."""
        # Extract the types of instance fields for this dataclass
        field_types = typing.get_type_hints(cls)
        field_types = {
            k: v
            for k, v in field_types.items()
            if k in {f.name for f in dataclasses.fields(cls)}
        }

        # Create sub-dataclasses if needed
        for name, klass in field_types.items():
            if hasattr(klass, "from_dict"):
                d[name] = klass.from_dict(d[name])

        return cls(**d)

    def serialize(self) -> bytes:
        """Serializes dataclass to bytes."""
        return msgpack.dumps(self.to_dict())

    @classmethod
    def deserialize(cls: Type[T], data: bytes) -> T:
        """Deserializes dataclass from bytes."""
        return cls.from_dict(msgpack.loads(data))


@dataclass
class _TaskRequest(_DataclassSerializerMixin):
    """The full task request to be sent to the pod."""

    serialized_protocol: SerializedProtocol
    pod_identifiers: List[str]
    aes_key: bytes


@dataclass
class _EncryptedTaskRequest(_DataclassSerializerMixin):
    """Encrypted task request."""

    encrypted_request: bytes


@dataclass
class _SignedEncryptedTaskRequest(_DataclassSerializerMixin):
    """Encrypted and signed task request."""

    encrypted_request: bytes
    signature: bytes


@dataclass
class _TaskRequestMessage(_DataclassSerializerMixin):
    """Task request message to be sent to pod."""

    serialized_protocol: SerializedProtocol
    auth_type: str
    request: bytes



"""Contains the base abstract class for all transformations."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, List, Optional, Type, cast
from uuid import uuid4

import attr
import desert
from marshmallow import Schema

from bitfount.transformations.exceptions import TransformationRegistryError
from bitfount.transformations.utils import _MarshmallowYamlShim

TRANSFORMATION_REGISTRY: Dict[str, Type[Transformation]] = {}
"""Dictionary of all registered transformations."""


# Keyword-only dataclasses allow us to get around the issue of subclasses having
# non-default args following default args (i.e. name and output). Every subclass
# should be marked using the same.
@attr.dataclass(kw_only=True)
class Transformation:
    """The base class that all transformations inherit from.

    By default these will be `attr.dataclass`es and their schemas will be generated
    from this. If more manual control over the schema is needed you may create a nested
    `Schema` class on the transformation that will be used instead. This nested
    class must inherit from `TransformationSchema`.

    Args:
        name: The name of the transformation. If not provided a unique name
            will be generated from the class name.
        output: Whether or not this transformation should be included in the
            final output. Defaults to False.

    Raises:
        TransformationRegistryError: If the transformation name is already in use.
        TransformationRegistryError: If the transformation name hasn't been provided
            and the transformation is not registered.
    """

    # Non-serializable variables
    _registry_name: ClassVar[Optional[str]] = None

    # Serializable variables
    name: Optional[str] = None
    output: bool = False

    def __attrs_post_init__(self) -> None:
        if not self.name:
            self.name = self._gen_name()

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        # Register concrete transformations with the registry
        if cls._registry_name:
            registry_name = cls._registry_name.lower()
            # Fail if name is already registered
            if registry_name in TRANSFORMATION_REGISTRY:
                raise TransformationRegistryError(
                    f"A transformation is already registered with name "
                    f'"{registry_name}"'
                )
            TRANSFORMATION_REGISTRY[registry_name] = cls

    @classmethod
    def schema(cls) -> Schema:
        """Gets an instance of the Schema associated with this Transformation.

        Raises:
            TypeError: If the transformation doesn't have a `TransformationSchema` as
                the schema.
        """
        # If the transformation has a custom schema, use it
        if hasattr(cls, "_Schema"):
            if issubclass(cls._Schema, _TransformationSchema):  # type: ignore[attr-defined] # Reason: hasattr check above # noqa: B950
                return cast(_TransformationSchema, cls._Schema())  # type: ignore[attr-defined] # Reason: hasattr check above # noqa: B950
            else:
                raise TypeError(
                    f"Schema attribute for class {cls.__name__} must be a "
                    f"TransformationSchema instance."
                )
        # Otherwise generate one from desert
        else:
            return desert.schema(cls, meta={"render_module": _MarshmallowYamlShim})

    @classmethod
    def _gen_name(cls) -> str:
        """Generates a unique name for the transformation."""
        if not cls._registry_name:
            raise TransformationRegistryError(
                f'Transformation "{cls.__name__}" isn\'t registered; '
                f"can't generate name."
            )
        slugged_reg_name = "-".join(cls._registry_name.split())
        return f"{slugged_reg_name.lower()}_{uuid4().hex}"


class MultiColumnOutputTransformation(ABC):
    """Marks a Transformation class as producing multi-column output."""

    @property
    @abstractmethod
    def columns(self) -> List[str]:
        """Returns a list of the columns that will be output."""
        raise NotImplementedError


class _TransformationSchema(Schema):
    """A custom schema class for more involved transformations.

    If you are inheriting from this note that you will need to ensure all fields
    in the inheritance hierarchy are included in the inherited schema.
    """

    class Meta:
        render_module = _MarshmallowYamlShim



"""PyTorch implementations of the datafactory module contents."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Union

from bitfount.backends.pytorch.data.dataloaders import (
    PyTorchBitfountDataLoader,
    PyTorchIterableBitfountDataLoader,
    _BasePyTorchBitfountDataLoader,
)
from bitfount.backends.pytorch.data.datasets import (
    _PyTorchDataset,
    _PyTorchIterableDataset,
)
from bitfount.data.datafactory import _DataFactory
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource

if TYPE_CHECKING:
    from bitfount.data.datasets import _BaseBitfountDataset
    from bitfount.data.datasplitters import DatasetSplitter
    from bitfount.data.schema import TableSchema
    from bitfount.data.types import DataSplit, _SemanticTypeValue
    from bitfount.transformations.batch_operations import BatchTimeOperation


class _PyTorchDataFactory(_DataFactory):
    """A PyTorch-specific implementation of the DataFactory provider."""

    def create_dataloader(
        self,
        dataset: _BaseBitfountDataset,
        batch_size: Optional[int] = None,
    ) -> _BasePyTorchBitfountDataLoader:
        """See base class."""
        if isinstance(dataset, _PyTorchIterableDataset):
            return PyTorchIterableBitfountDataLoader(
                dataset=dataset, batch_size=batch_size
            )
        elif isinstance(dataset, _PyTorchDataset):
            return PyTorchBitfountDataLoader(dataset, batch_size=batch_size)

        raise TypeError(
            "The _PyTorchDataFactory class only supports "
            "subclasses of PyTorch Dataset for creating a DataLoader."
        )

    def create_dataset(
        self,
        datasource: BaseSource,
        data_splitter: Optional[DatasetSplitter],
        data_split: DataSplit,
        schema: TableSchema,
        selected_cols: List[str],
        selected_cols_semantic_types: Mapping[_SemanticTypeValue, List[str]],
        target: Optional[Union[str, List[str]]] = None,
        batch_transforms: Optional[List[BatchTimeOperation]] = None,
        **kwargs: Any,
    ) -> Union[_PyTorchDataset, _PyTorchIterableDataset]:
        """See base class."""
        if datasource.multi_table and isinstance(datasource, DatabaseSource):
            return _PyTorchIterableDataset(
                schema=schema,
                selected_cols_semantic_types=selected_cols_semantic_types,
                data_splitter=data_splitter,
                datasource=datasource,
                target=target,
                selected_cols=selected_cols,
                batch_transforms=batch_transforms,
                data_split=data_split,
                **kwargs,
            )

        return _PyTorchDataset(
            schema=schema,
            selected_cols_semantic_types=selected_cols_semantic_types,
            data_splitter=data_splitter,
            datasource=datasource,
            target=target,
            selected_cols=selected_cols,
            batch_transforms=batch_transforms,
            data_split=data_split,
            **kwargs,
        )
